"""Custom format checkers for jsonschema validation."""

import re

import astropy.units as u
import jsonschema

from simtools.utils import names

format_checker = jsonschema.FormatChecker()


@format_checker.checks("astropy_unit")
def check_astropy_unit(unit_string):
    """Validate astropy units (including dimensionless) for jsonschema."""
    try:
        u.Unit(unit_string)
    except (ValueError, TypeError) as exc:
        if unit_string != "dimensionless":
            raise ValueError(f"'{unit_string}' is not a valid Unit") from exc
    return True


@format_checker.checks("astropy_unit_of_time")
def check_astropy_unit_of_time(unit_string):
    """Validate astropy units that this is an astropy unit of time."""
    u.Unit(unit_string).to("s")
    return True


@format_checker.checks("astropy_unit_of_length)")
def check_astropy_unit_of_length(unit_string):
    """Validate astropy units that this is an astropy unit of length."""
    u.Unit(unit_string).to("m")
    return True


@format_checker.checks("array_element")
def check_array_element(element):
    """Validate array elements for jsonschema."""
    names.validate_array_element_name(element)
    return True


@format_checker.checks("array_triggers_name")
def check_array_triggers_name(name):
    """Validate array trigger names for jsonschema."""
    pattern = r"(.*)(?=_single_telescope|_array)"
    if not re.match(pattern, name):
        raise ValueError(f"Array trigger name '{name}' does not match pattern '{pattern}'")
    names.validate_array_element_type(re.match(pattern, name).group(1))
    return True



"""Calculate the thresholds for energy, radial distance, and viewcone."""

import astropy.units as u
import matplotlib.pyplot as plt
import numpy as np
import tables
from astropy.coordinates import AltAz
from ctapipe.coordinates import GroundFrame, TiltedGroundFrame


class LimitCalculator:
    """
    Compute thresholds/limits for energy, radial distance, and viewcone.

    Event data is read from the reduced MC event data file.

    Parameters
    ----------
    event_data_file : str
        Path to the HDF5 file containing the event data.
    telescope_list : list, optional
        List of telescope IDs to filter the events (default is None).
    """

    def __init__(self, event_data_file, telescope_list=None):
        """
        Initialize the LimitCalculator with the given event data file.

        Parameters
        ----------
        event_data_file : str
            Path to the reduced MC event data file.
        telescope_list : list, optional
            List of telescope IDs to filter the events (default is None).
        """
        self.event_data_file = event_data_file
        self.telescope_list = telescope_list
        self.event_x_core = None
        self.event_y_core = None
        self.simulated = None
        self.shower_id_triggered = None
        self.list_of_files = None
        self.shower_sim_azimuth = None
        self.shower_sim_altitude = None
        self.array_azimuth = None
        self.array_altitude = None
        self.trigger_telescope_list_list = None
        self.units = {}
        self._read_event_data()

    def _read_event_data(self):
        """Read the event data from the reduced MC event data file."""
        with tables.open_file(self.event_data_file, mode="r") as f:
            reduced_data = f.root.data.reduced_data
            triggered_data = f.root.data.triggered_data
            file_names = f.root.data.file_names
            trigger_telescope_list_list = f.root.data.trigger_telescope_list_list

            self.event_x_core = reduced_data.col("core_x")
            self.event_y_core = reduced_data.col("core_y")
            self.simulated = reduced_data.col("simulated")
            self.shower_id_triggered = triggered_data.col("shower_id_triggered")
            self.list_of_files = file_names.col("file_names")
            self.shower_sim_azimuth = reduced_data.col("shower_sim_azimuth")
            self.shower_sim_altitude = reduced_data.col("shower_sim_altitude")
            self.array_altitude = reduced_data.col("array_altitudes")
            self.array_azimuth = reduced_data.col("array_azimuths")

            self.trigger_telescope_list_list = [
                [np.int16(tel) for tel in event] for event in trigger_telescope_list_list
            ]

    def _compute_limits(self, hist, bin_edges, loss_fraction, limit_type="lower"):
        """
        Compute the limits based on the loss fraction.

        Parameters
        ----------
        hist : np.ndarray
            1D histogram array.
        bin_edges : np.ndarray
            Array of bin edges.
        loss_fraction : float
            Fraction of events to be lost.
        limit_type : str, optional
            Type of limit ('lower' or 'upper'). Default is 'lower'.

        Returns
        -------
        float
            Bin edge value corresponding to the threshold.
        """
        cumulative_sum = np.cumsum(hist) if limit_type == "upper" else np.cumsum(hist[::-1])
        total_events = np.sum(hist)
        threshold = (1 - loss_fraction) * total_events
        bin_index = np.searchsorted(cumulative_sum, threshold)
        return bin_edges[bin_index] if limit_type == "upper" else bin_edges[-bin_index]

    def _prepare_data_for_limits(self):
        """
        Prepare the data required for computing limits.

        Returns
        -------
        tuple
            Tuple containing core distances, triggered energies, core bins, and energy bins.
        """
        shower_id_triggered_masked = self.shower_id_triggered
        if self.telescope_list is not None:
            mask = np.array(
                [
                    all(tel in event for tel in self.telescope_list)
                    for event in self.trigger_telescope_list_list
                ]
            )
            shower_id_triggered_masked = self.shower_id_triggered[mask]

        triggered_energies = self.simulated[shower_id_triggered_masked]
        energy_bins = np.logspace(
            np.log10(triggered_energies.min()), np.log10(triggered_energies.max()), 1000
        )
        event_x_core_shower, event_y_core_shower = self._transform_to_shower_coordinates()
        core_distances_all = np.sqrt(event_x_core_shower**2 + event_y_core_shower**2)
        core_distances_triggered = core_distances_all[shower_id_triggered_masked]
        core_bins = np.linspace(
            core_distances_triggered.min(), core_distances_triggered.max(), 1000
        )

        return core_distances_triggered, triggered_energies, core_bins, energy_bins

    def compute_lower_energy_limit(self, loss_fraction):
        """
        Compute the lower energy limit in TeV based on the event loss fraction.

        Parameters
        ----------
        loss_fraction : float
            Fraction of events to be lost.

        Returns
        -------
        astropy.units.Quantity
            Lower energy limit.
        """
        _, triggered_energies, _, energy_bins = self._prepare_data_for_limits()

        hist, _ = np.histogram(triggered_energies, bins=energy_bins)
        lower_bin_edge_value = self._compute_limits(
            hist, energy_bins, loss_fraction, limit_type="lower"
        )
        return lower_bin_edge_value * u.TeV

    def compute_upper_radial_distance(self, loss_fraction):
        """
        Compute the upper radial distance based on the event loss fraction.

        Parameters
        ----------
        loss_fraction : float
            Fraction of events to be lost.

        Returns
        -------
        astropy.units.Quantity
            Upper radial distance in m.
        """
        core_distances_triggered, _, core_bins, _ = self._prepare_data_for_limits()

        hist, _ = np.histogram(core_distances_triggered, bins=core_bins)
        upper_bin_edge_value = self._compute_limits(
            hist, core_bins, loss_fraction, limit_type="upper"
        )
        return upper_bin_edge_value * u.m

    def compute_viewcone(self, loss_fraction):
        """
        Compute the viewcone based on the event loss fraction.

        Parameters
        ----------
        loss_fraction : float
            Fraction of events to be lost.

        Returns
        -------
        astropy.units.Quantity
            Viewcone radius in degrees.
        """
        # already in radians
        azimuth_diff = self.array_azimuth - self.shower_sim_azimuth  # * (np.pi / 180.0)
        sim_altitude_rad = self.shower_sim_altitude  # * (np.pi / 180.0)
        array_altitude_rad = self.array_altitude  # * (np.pi / 180.0)
        x_1 = np.cos(azimuth_diff) * np.cos(sim_altitude_rad)
        y_1 = np.sin(azimuth_diff) * np.cos(sim_altitude_rad)
        z_1 = np.sin(sim_altitude_rad)
        x_2 = x_1 * np.sin(array_altitude_rad) - z_1 * np.cos(array_altitude_rad)
        y_2 = y_1
        z_2 = x_1 * np.cos(array_altitude_rad) + z_1 * np.sin(array_altitude_rad)
        off_angles = np.arctan2(np.sqrt(x_2**2 + y_2**2), z_2) * (180.0 / np.pi)
        angle_bins = np.linspace(off_angles.min(), off_angles.max(), 400)
        hist, _ = np.histogram(off_angles, bins=angle_bins)

        upper_bin_edge_value = self._compute_limits(
            hist, angle_bins, loss_fraction, limit_type="upper"
        )
        return upper_bin_edge_value * u.deg

    def _transform_to_shower_coordinates(self):
        """
        Transform core positions from ground coordinates to shower coordinates.

        Returns
        -------
        tuple
            Core positions in shower coordinates (x, y).
        """
        pointing_az = self.shower_sim_azimuth * u.rad
        pointing_alt = self.shower_sim_altitude * u.rad

        pointing = AltAz(az=pointing_az, alt=pointing_alt)
        ground = GroundFrame(x=self.event_x_core * u.m, y=self.event_y_core * u.m, z=0 * u.m)
        shower_frame = ground.transform_to(TiltedGroundFrame(pointing_direction=pointing))

        return shower_frame.x.value, shower_frame.y.value

    def plot_data(self):
        """Plot the core distances and energies of triggered events."""
        shower_id_triggered_masked = self.shower_id_triggered
        if self.telescope_list is not None:
            mask = np.array(
                [
                    all(tel in event for tel in self.telescope_list)
                    for event in self.trigger_telescope_list_list
                ]
            )
            shower_id_triggered_masked = self.shower_id_triggered[mask]

        core_distances_all = np.sqrt(self.event_x_core**2 + self.event_y_core**2)
        core_distances_triggered = core_distances_all[shower_id_triggered_masked]
        triggered_energies = self.simulated[shower_id_triggered_masked]

        core_bins = np.linspace(core_distances_triggered.min(), core_distances_triggered.max(), 400)
        energy_bins = np.logspace(
            np.log10(triggered_energies.min()), np.log10(triggered_energies.max()), 400
        )
        plt.figure(figsize=(8, 6))
        plt.hist2d(
            core_distances_triggered,
            triggered_energies,
            bins=[core_bins, energy_bins],
            norm="log",
            cmap="viridis",
        )

        plt.colorbar(label="Event Count")
        plt.xlabel("Core Distance [m]")
        plt.ylabel("Energy [TeV]")
        plt.yscale("log")
        plt.title("2D Histogram of Triggered Core Distance vs Energy")
        plt.show()



#!/usr/bin/python3

r"""
    Write sim_telarray histograms into pdf and hdf5 files.

    It accepts multiple lists of histograms files, a single list or a histogram file.
    Each histogram is plotted in a page of the pdf file if the --pdf option is activated.


    Command line arguments
    ----------------------
    hist_file_names (str, optional)
        Name of the histogram files to be plotted.
        It can be given as the histogram file names (more than one option allowed) or as a text
        file with the names of the histogram files in it.
    pdf (bool, optional)
        If set, histograms are saved into pdf files.
        One pdf file contains all the histograms found in the file.
        The name of the file is controlled via output_file_name.
    hdf5: bool
        If true, histograms are saved into hdf5 files.
        At least one of pdf and hdf5 has to be activated.
    output_file_name (str, optional)
        The name of the output hdf5 (and/or pdf) files (without the path).
        If not given, output_file_name takes the name from the (first) input file
        (hist_file_names).
        If the output output_file_name.hdf5 file already exists and hdf5 is set, the tables
        associated to hdf5 will be overwritten. The remaining tables, if any, will stay
        untouched.
    test: bool
        Test option. Generate only two histograms for testing purposes.

    Raises
    ------
    TypeError:
        if argument passed through hist_file_names is not a file.

    Example
    -------
    .. code-block:: console

        simtools-generate-simtel-array-histograms --hist_file_names tests/resources/ \\
            run2_gamma_za20deg_azm0deg-North-Prod5_test-production-5.hdata.zst \\
            --output_file_name test_hist_hdata --hdf5 --pdf

"""

import logging
from pathlib import Path

import matplotlib.pyplot as plt
from matplotlib.backends.backend_pdf import PdfPages

import simtools.utils.general as gen
from simtools.configuration import configurator
from simtools.io_operations import io_handler
from simtools.simtel.simtel_io_histograms import SimtelIOHistograms


def _parse(label, description):
    """
    Parse command line configuration.

    Parameters
    ----------
    label: str
        Label describing the application.
    description: str
        Description of the application.

    Returns
    -------
    CommandLineParser
        Command line parser object

    """
    config = configurator.Configurator(label=label, description=description)

    config.parser.add_argument(
        "--hist_file_names",
        help="Name of the histogram files to be plotted or the text file containing the list of "
        "histogram files.",
        nargs="+",
        required=True,
        type=str,
    )

    config.parser.add_argument(
        "--hdf5", help="Save histograms into a hdf5 file.", action="store_true", required=False
    )

    config.parser.add_argument(
        "--pdf", help="Save histograms into a pdf file.", action="store_true", required=False
    )

    config.parser.add_argument(
        "--output_file_name",
        help="Name of the hdf5 (and/or pdf) file where to save the histograms.",
        type=str,
        required=False,
        default=None,
    )

    config_parser, _ = config.initialize(db_config=False, paths=True)
    if not config_parser["pdf"] and not config_parser["hdf5"]:
        config.parser.error("At least one argument is required: --pdf or --hdf5.")

    return config_parser


def check_and_log_overwrite(config_parser, logger):
    """
    Check if the output hdf5 file already exists and log a warning if it does.

    Parameters
    ----------
    config_parser: dict
        Parsed command line arguments.
    logger: logging.Logger
        Logger object for logging messages.

    Returns
    -------
    bool
        True if the hdf5 file exists and should be overwritten.
    """
    if Path(f"{config_parser['output_file_name']}.hdf5").exists() and config_parser["hdf5"]:
        msg = (
            f"Output hdf5 file {config_parser['output_file_name']}.hdf5 already exists. "
            f"Overwriting it."
        )
        logger.warning(msg)
        return True
    return False


def create_pdf(simtel_histograms, output_file_name, config_parser, logger):
    """
    Create a PDF file containing histograms.

    Parameters
    ----------
    simtel_histograms: SimtelIOHistograms
        SimtelIOHistograms object containing histograms to plot.
    output_file_name: str
        Base name for the output PDF file.
    config_parser: dict
        Parsed command line arguments.
    logger: logging.Logger
        Logger object for logging messages.
    """
    if config_parser["pdf"]:
        logger.debug(f"Creating the pdf file {output_file_name}.pdf")
        pdf_pages = PdfPages(f"{output_file_name}.pdf")
        number_of_histograms = 2 if config_parser["test"] else len(simtel_histograms.combined_hists)
        for i_hist in range(number_of_histograms):
            logger.debug(f"Processing: {i_hist + 1} histogram.")
            fig, ax = plt.subplots(1, 1, figsize=(6, 6))
            simtel_histograms.plot_one_histogram(i_hist, ax)
            plt.tight_layout()
            pdf_pages.savefig(fig)
            plt.clf()
        plt.close()
        pdf_pages.close()
        logger.info(f"Wrote histograms to the pdf file {output_file_name}.pdf")


def main():  # noqa: D103
    label = Path(__file__).stem
    description = "Display simtel_array histograms and/or write them into hdf5 format."
    io_handler_instance = io_handler.IOHandler()
    config_parser = _parse(label, description)
    output_path = io_handler_instance.get_output_directory(label, sub_dir="application-plots")
    logger = logging.getLogger()
    logger.setLevel(gen.get_log_level_from_user(config_parser["log_level"]))

    histogram_files = gen.get_list_of_files_from_command_line(
        config_parser["hist_file_names"], [".zst", ".simtel", ".hdata"]
    )

    # If no output name is passed, the tool gets the name of the first histogram of the list
    if config_parser["output_file_name"] is None:
        config_parser["output_file_name"] = Path(histogram_files[0]).absolute().name
    output_file_name = Path(output_path).joinpath(f"{config_parser['output_file_name']}")

    simtel_histograms = SimtelIOHistograms(histogram_files)
    create_pdf(simtel_histograms, output_file_name, config_parser, logger)
    if config_parser["hdf5"]:
        simtel_histograms.export_histograms(
            f"{output_file_name}.hdf5",
            overwrite=check_and_log_overwrite(config_parser, logger),
        )


if __name__ == "__main__":
    main()



"""HT Condor script generator for simulation production."""

import logging
from pathlib import Path

import astropy.units as u

_logger = logging.getLogger(__name__)


def generate_submission_script(args_dict):
    """
    Generate the HT Condor submission script.

    Parameters
    ----------
    args_dict: dict
        Arguments dictionary.
    """
    _logger.info("Generating HT Condor submission scripts ")

    work_dir = Path(args_dict["output_path"])
    log_dir = work_dir / "logs"
    work_dir.mkdir(parents=True, exist_ok=True)
    log_dir.mkdir(parents=True, exist_ok=True)
    submit_file_name = "simulate_prod.submit"

    with open(work_dir / f"{submit_file_name}.condor", "w", encoding="utf-8") as submit_file_handle:
        submit_file_handle.write(
            _get_submit_file(
                f"{submit_file_name}.sh",
                args_dict["apptainer_image"],
                args_dict["priority"],
                +args_dict["number_of_runs"],
            )
        )

    with open(work_dir / f"{submit_file_name}.sh", "w", encoding="utf-8") as submit_script_handle:
        submit_script_handle.write(_get_submit_script(args_dict))

    Path(work_dir / f"{submit_file_name}.sh").chmod(0o755)


def _get_submit_file(executable, apptainer_image, priority, n_jobs):
    """
    Return HT Condor submit file.

    Database access variables are passed through the environment file.

    Parameters
    ----------
    executable: str
        Name of the executable script.
    apptainer_image: str
        Path to the Apptainer image.
    priority: int
        Priority of the job.
    n_jobs: int
        Number of jobs to queue.

    Returns
    -------
    str
        HT Condor submit file content.
    """
    return f"""universe = container
container_image = {apptainer_image}
transfer_container = false

executable = {executable}
error      = logs/err.$(cluster)_$(process)
output     = logs/out.$(cluster)_$(process)
log        = logs/log.$(cluster)_$(process)

priority = {priority}
arguments = "$(process) env.txt"

queue {n_jobs}
"""


def _get_submit_script(args_dict):
    """
    Return HT Condor submit script.

    Parameters
    ----------
    args_dict: dict
        Arguments dictionary.

    Returns
    -------
    str
        HT Condor submit script content.
    """
    azimuth_angle_string = f"{args_dict['azimuth_angle'].to(u.deg).value}"
    zenith_angle_string = f"{args_dict['zenith_angle'].to(u.deg).value}"
    energy_range = args_dict["energy_range"]
    energy_range_string = (
        f'"{energy_range[0].to(u.GeV).value} GeV {energy_range[1].to(u.GeV).value} GeV"'
    )
    core_scatter = args_dict["core_scatter"]
    core_scatter_string = f'"{core_scatter[0]} {core_scatter[1].to(u.m).value} m"'

    label = args_dict["label"] if args_dict["label"] else "simulate-prod"

    return f"""#!/usr/bin/env bash

# Process ID used to generate run number
process_id="$1"
# Load environment variables (for DB access)
set -a; source "$2"

simtools-simulate-prod \\
    --simulation_software {args_dict["simulation_software"]} \\
    --label {label} \\
    --model_version {args_dict["model_version"]} \\
    --site {args_dict["site"]} \\
    --array_layout_name {args_dict["array_layout_name"]} \\
    --primary {args_dict["primary"]} \\
    --azimuth_angle {azimuth_angle_string} \\
    --zenith_angle {zenith_angle_string} \\
    --nshow {args_dict["nshow"]} \\
    --energy_range {energy_range_string} \\
    --core_scatter {core_scatter_string} \\
    --run_number_start $((process_id + {args_dict["run_number_start"]})) \\
    --number_of_runs 1 \\
    --submit_engine \"local\" \\
    --data_directory /tmp/simtools-data \\
    --output_path /tmp/simtools-output \\
    --log_level {args_dict["log_level"]} \\
    --pack_for_grid_register simtools-output
"""



#!/usr/bin/python3

"""
Inspect databases and print (available database names and collections).

Command line arguments
----------------------
db_name (str, optional)
    Inspect a specific database.
"""

import logging

import simtools.utils.general as gen
from simtools.configuration import configurator
from simtools.db import db_handler


def _parse():
    config = configurator.Configurator(description="Inspect databases")
    config.parser.add_argument(
        "--db_name",
        help="Inspect a specific database (use all to print all databases)",
        default="all",
        required=True,
    )
    return config.initialize(db_config=True, simulation_model="telescope")


def main():  # noqa: D103
    args_dict, db_config = _parse()

    logger = logging.getLogger()
    logger.setLevel(gen.get_log_level_from_user(args_dict["log_level"]))

    db = db_handler.DatabaseHandler(mongo_db_config=db_config)

    databases = db.db_client.list_database_names()

    for db_name in databases:
        if args_dict["db_name"] != "all" and db_name != args_dict["db_name"]:
            continue
        # missing admin rights; skip config and admin
        if db_name in ("config", "admin", "local"):
            continue
        print("Database:", db_name)
        collections = db.get_collections(db_name=db_name)
        print("   Collections:", collections)


if __name__ == "__main__":
    main()



#!/usr/bin/python3

"""
    Simulate the cumulative PSF and compare with data (if available).

    The telescope zenith angle and the source distance can be set by command line arguments.

    The measured cumulative PSF should be provided by using the command line argument data. \
    A file name is expected, in which the file should contains 3 columns: radial distance in mm, \
    differential value of photon intensity and its integral value.

    The MC model can be changed by providing a yaml file with the new parameter values using \
    the argument pars (see example below).

    Examples of the plots generated by this applications are shown below. On the left, \
    the cumulative PSF and on the right, the simulated PSF image.

    .. _validate_cumulative_psf_plot:
    .. image::  images/validate_cumulative_psf_North-LST-1_cumulativePSF.png
      :width: 49 %
    .. image::  images/validate_cumulative_psf_North-LST-1_image.png
      :width: 49 %

    Command line arguments
    ----------------------
    site (str, required)
        North or South.
    telescope (str, required)
        Telescope model name (e.g. LST-1, SST-D, ...).
    model_version (str, optional)
        Model version.
    src_distance (float, optional)
        Source distance in km.
    zenith (float, optional)
        Zenith angle in deg.
    data (str, optional)
        Name of the data file with the measured cumulative PSF.
    pars (str, optional)
        Yaml file with the new model parameters to replace the default ones.
    test (activation mode, optional)
        If activated, application will be faster by simulating fewer photons.

    Example
    -------
    LST-1 5.0.0

    Runtime < 1 min.

    Get an example dataset from the DB:

    .. code-block:: console

        simtools-get-file-from-db --file_name PSFcurve_data_v2.txt

    Run the application:

    .. code-block:: console

        simtools-compare-cumulative-psf --site North --telescope LST-1 \
            --model_version 5.0.0 --data PSFcurve_data_v2.txt

    The output is saved in simtools-output/validate_cumulative_psf

    Expected final print-out message:

    .. code-block:: console

        d80 in cm = 3.3662565358159013

"""

import logging
from collections import OrderedDict
from pathlib import Path

import astropy.units as u
import matplotlib.pyplot as plt
import numpy as np

import simtools.utils.general as gen
from simtools.configuration import configurator
from simtools.io_operations import io_handler
from simtools.model.telescope_model import TelescopeModel
from simtools.ray_tracing.ray_tracing import RayTracing
from simtools.visualization import visualize


def _parse(label):
    config = configurator.Configurator(
        label=label,
        description=(
            "Calculate and plot the PSF and eff. mirror area as a function of off-axis angle "
            "of the telescope requested."
        ),
    )
    config.parser.add_argument(
        "--src_distance",
        help="Source distance in km",
        type=float,
        default=10,
    )
    config.parser.add_argument(
        "--zenith",
        help="Zenith angle in deg",
        type=float,
        default=20.0,
    )
    config.parser.add_argument(
        "--data",
        help="Data file name with the measured PSF vs radius [cm]",
        type=str,
    )
    return config.initialize(db_config=True, simulation_model=["telescope", "model_version"])


def load_data(datafile):
    """Load the data file with the measured PSF vs radius [cm]."""
    radius_cm = "Radius [cm]"
    relative_intensity = "Relative intensity"

    d_type = {"names": (radius_cm, relative_intensity), "formats": ("f8", "f8")}
    data = np.loadtxt(datafile, dtype=d_type, usecols=(0, 2))
    data[radius_cm] *= 0.1
    data[relative_intensity] /= np.max(np.abs(data[relative_intensity]))
    return data


def main():  # noqa: D103
    label = Path(__file__).stem
    args_dict, db_config = _parse(label)

    logger = logging.getLogger()
    logger.setLevel(gen.get_log_level_from_user(args_dict["log_level"]))

    # Output directory to save files related directly to this app
    _io_handler = io_handler.IOHandler()
    output_dir = _io_handler.get_output_directory(label, sub_dir="application-plots")

    tel_model = TelescopeModel(
        site=args_dict["site"],
        telescope_name=args_dict["telescope"],
        mongo_db_config=db_config,
        model_version=args_dict["model_version"],
        label=label,
    )
    if args_dict.get("telescope_model_file"):
        tel_model.change_multiple_parameters_from_file(args_dict["telescope_model_file"])

    ray = RayTracing(
        telescope_model=tel_model,
        simtel_path=args_dict["simtel_path"],
        zenith_angle=args_dict["zenith"] * u.deg,
        source_distance=args_dict["src_distance"] * u.km,
        off_axis_angle=[0.0] * u.deg,
    )

    ray.simulate(test=args_dict["test"], force=False)
    ray.analyze(force=False)

    # Plotting cumulative PSF
    im = ray.images()[0]

    print(f"d80 in cm = {im.get_psf()}")

    # Plotting cumulative PSF
    # Measured cumulative PSF
    data_to_plot = OrderedDict()
    radius = None
    if args_dict.get("data", None):
        data_file = gen.find_file(args_dict["data"], args_dict["model_path"])
        data_to_plot["measured"] = load_data(data_file)
        radius = data_to_plot["measured"]["Radius [cm]"]

    # Simulated cumulative PSF
    if radius is not None:
        data_to_plot[r"sim$\_$telarray"] = im.get_cumulative_data(radius * u.cm)
    else:
        raise ValueError("Radius data is not available. Cannot compute cumulative PSF.")

    fig = visualize.plot_1d(data_to_plot)
    fig.gca().set_ylim(0, 1.05)

    plot_file_name = label + "_" + tel_model.name + "_cumulative_PSF"
    plot_file = output_dir.joinpath(plot_file_name)
    visualize.save_figure(fig, plot_file)

    # Plotting image
    data_to_plot = im.get_image_data()
    fig = visualize.plot_hist_2d(data_to_plot, bins=80)
    circle = plt.Circle((0, 0), im.get_psf(0.8) / 2, color="k", fill=False, lw=2, ls="--")
    fig.gca().add_artist(circle)
    fig.gca().set_aspect("equal")

    plot_file_name = label + "_" + tel_model.name + "_image"
    plot_file = output_dir.joinpath(plot_file_name)
    visualize.save_figure(fig, plot_file)


if __name__ == "__main__":
    main()



"""Helper functions for integration testing."""

import os
from pathlib import Path


def skip_camera_efficiency(config):
    """Skip camera efficiency tests if the old version of testeff is used."""
    if "camera-efficiency" in config["APPLICATION"] and not _new_testeff_version():
        return (
            "Any applications calling the old version of testeff are skipped "
            "due to a limitation of the old testeff not allowing to specify "
            "the include directory. Please update your sim_telarray tarball."
        )
    return None


def _new_testeff_version():
    """
    Testeff has been updated to allow to specify the include directory.

    This test checks if the new version is used.
    """
    testeff_path = Path(os.getenv("SIMTOOLS_SIMTEL_PATH")) / "sim_telarray/testeff.c"
    try:
        with open(testeff_path, encoding="utf-8") as file:
            file_content = file.read()
            if (
                "/* Combine the include paths such that those from '-I...' options */"
                in file_content
            ):
                return True
            return False
    except FileNotFoundError as exc:
        raise FileNotFoundError("The testeff executable could not be found.") from exc



"""Functions asserting certain conditions are met (used e.g., in integration tests)."""

import json
import logging
from collections import defaultdict
from pathlib import Path

import numpy as np
import yaml

_logger = logging.getLogger(__name__)


def assert_file_type(file_type, file_name):
    """
    Assert that the file is of the given type.

    Parameters
    ----------
    file_type: str
        File type (json, yaml).
    file_name: str
        File name.

    """
    if file_type == "json":
        try:
            with open(file_name, encoding="utf-8") as file:
                json.load(file)
            return True
        except (json.JSONDecodeError, FileNotFoundError):
            return False
    if file_type in ("yaml", "yml"):
        if Path(file_name).suffix[1:] not in ("yaml", "yml"):
            return False
        try:
            with open(file_name, encoding="utf-8") as file:
                yaml.safe_load(file)
            return True
        except (yaml.YAMLError, FileNotFoundError):
            return False

    # no dedicated tests for other file types, checking suffix only
    _logger.info(f"File type test is checking suffix only for {file_name} (suffix: {file_type}))")
    return Path(file_name).suffix[1:] == file_type


def assert_n_showers_and_energy_range(file):
    """
    Assert the number of showers and the energy range.

    The number of showers should be consistent with the required one (up to 1% tolerance)
    and the energies simulated are required to be within the configured ones.

    Parameters
    ----------
    file: Path
        Path to the sim_telarray file.

    """
    from eventio.simtel.simtelfile import SimTelFile  # pylint: disable=import-outside-toplevel

    simulated_energies = []
    simulation_config = {}
    with SimTelFile(file) as f:
        simulation_config = f.mc_run_headers[0]
        for event in f.iter_mc_events():
            simulated_energies.append(event["mc_shower"]["energy"])

    # The relative tolerance is set to 1% because ~0.5% shower simulations do not
    # succeed, without resulting in an error. This tolerance therefore is not an issue.
    consistent_n_showers = np.isclose(
        len(np.unique(simulated_energies)), simulation_config["n_showers"], rtol=1e-2
    )
    consistent_energy_range = all(
        simulation_config["E_range"][0] <= energy <= simulation_config["E_range"][1]
        for energy in simulated_energies
    )

    return consistent_n_showers and consistent_energy_range


def assert_expected_output(file, expected_output):
    """
    Assert that the expected output is present in the sim_telarray file.

    Parameters
    ----------
    file: Path
        Path to the sim_telarray file.
    expected_output: dict
        Expected output values.

    """
    from eventio.simtel.simtelfile import SimTelFile  # pylint: disable=import-outside-toplevel

    item_to_check = defaultdict(list)
    with SimTelFile(file) as f:
        for event in f:
            if "pe_sum" in expected_output:
                item_to_check["pe_sum"].extend(
                    event["photoelectron_sums"]["n_pe"][event["photoelectron_sums"]["n_pe"] > 0]
                )
            if "trigger_time" in expected_output:
                item_to_check["trigger_time"].extend(event["trigger_information"]["trigger_times"])
            if "photons" in expected_output:
                item_to_check["photons"].extend(
                    event["photoelectron_sums"]["photons_atm_qe"][
                        event["photoelectron_sums"]["photons"] > 0
                    ]
                )

    for key, value in expected_output.items():
        if len(item_to_check[key]) == 0:
            _logger.error(f"No data found for {key}")
            return False

        if not value[0] < np.mean(item_to_check[key]) < value[1]:
            _logger.error(
                f"Mean of {key} is not in the expected range, got {np.mean(item_to_check[key])}"
            )
            return False

    return True


def check_output_from_sim_telarray(file, expected_output):
    """
    Check that the sim_telarray simulation result is reasonable and matches the expected output.

    Parameters
    ----------
    file: Path
        Path to the sim_telarray file.
    expected_output: dict
        Expected output values.

    Raises
    ------
    ValueError
        If the file is not a zstd compressed file.
    """
    if file.suffix != ".zst":
        raise ValueError(
            f"Expected output file {file} is not a zstd compressed file "
            f"(i.e., a sim_telarray file)."
        )

    return assert_n_showers_and_energy_range(file=file) and assert_expected_output(
        file=file, expected_output=expected_output
    )



"""Helper functions for legend handlers used for plotting."""

import matplotlib.colors as mcolors
import matplotlib.patches as mpatches
import numpy as np

__all__ = [
    "EdgePixelObject",
    "HexEdgePixelHandler",
    "HexOffPixelHandler",
    "HexPixelHandler",
    "LSTHandler",
    "LSTObject",
    "MSTHandler",
    "MSTObject",
    "MeanRadiusOuterEdgeHandler",
    "MeanRadiusOuterEdgeObject",
    "OffPixelObject",
    "PixelObject",
    "SCTHandler",
    "SCTObject",
    "SSTHandler",
    "SSTObject",
    "SquareEdgePixelHandler",
    "SquareOffPixelHandler",
    "SquarePixelHandler",
    "TelescopeHandler",
]

# assume same size of telescope at both sides
# (good assumption for plotting)
lst = "LSTN"
mst = "MSTN"
sst = "SSTS"
sct = "SCTS"
hess = "HESS"
magic = "MAGIC"
veritas = "VERITAS"


def calculate_center(handlebox, width_factor=3, height_factor=3):
    """
    Calculate the center of the handlebox based on given factors.

    Parameters
    ----------
    handlebox: matplotlib.legend.Legend
        The handlebox object from the legend.
    width_factor: int, optional
        The factor to adjust the width.
    height_factor: int, optional
        The factor to adjust the height.

    Returns
    -------
    tuple
        The calculated (x0, y0) position.
    """
    x0 = handlebox.xdescent + handlebox.width / width_factor
    y0 = handlebox.ydescent + handlebox.height / height_factor
    return x0, y0


class TelescopeHandler:
    """
    Telescope handler that centralizes the telescope information.

    Individual telescopes handlers inherit from this class.
    """

    def __init__(self, radius=None):
        self.colors_dict = {
            "LSTN": "darkorange",
            "MSTx": "dodgerblue",
            "MSTN": "dodgerblue",
            "LSTS": "darkorange",
            "MSTS": "dodgerblue",
            "SCTS": "black",
            "SSTS": "darkgreen",
            "HESS": "grey",
            "MAGIC": "grey",
            "VERITAS": "grey",
        }

        # hardwired values; this is for plotting purposes only
        self.radius_dict = {
            "LSTN": 12.5,
            "MSTx": 9.15,
            "MSTN": 9.15,
            "LSTS": 12.5,
            "MSTS": 9.15,
            "SCTS": 7.15,
            "SSTS": 3.0,
            "HESS": 6.0,
            "MAGIC": 8.5,
            "VERITAS": 6.0,
        }


class PixelObject:
    """Pixel Object."""


class EdgePixelObject:
    """Edge-Pixel Object."""


class OffPixelObject:
    """Off-Pixel Object."""


class LSTObject:
    """LST Object."""


class MSTObject:
    """MST Object."""


class SCTObject:
    """SCT Object."""


class SSTObject:
    """SST Object."""


class HESSObject:
    """HESS Object."""


class MAGICObject:
    """MAGIC Object."""


class VERITASObject:
    """VERITAS Object."""


class MeanRadiusOuterEdgeObject:
    """Object for Mean radius outer edge."""


class HexPixelHandler:
    """Legend handler class to plot a hexagonal "on" pixel."""

    @staticmethod
    def legend_artist(_, __, ___, handlebox):  # noqa: D102
        x0, y0 = calculate_center(handlebox)
        patch = mpatches.RegularPolygon(
            (x0, y0),
            numVertices=6,
            radius=0.7 * handlebox.height,
            orientation=np.deg2rad(30),
            facecolor=(1, 1, 1, 0),
            edgecolor=(0, 0, 0, 1),
            transform=handlebox.get_transform(),
        )
        handlebox.add_artist(patch)
        return patch


class HexEdgePixelHandler:
    """Legend handler class to plot a hexagonal "edge" pixel."""

    @staticmethod
    def legend_artist(_, __, ___, handlebox):  # noqa: D102
        x0, y0 = calculate_center(handlebox)
        patch = mpatches.RegularPolygon(
            (x0, y0),
            numVertices=6,
            radius=0.7 * handlebox.height,
            orientation=np.deg2rad(30),
            facecolor=(*mcolors.to_rgb("brown"), 0.5),
            edgecolor=(*mcolors.to_rgb("black"), 1),
            transform=handlebox.get_transform(),
        )
        handlebox.add_artist(patch)
        return patch


class HexOffPixelHandler:
    """Legend handler class to plot a hexagonal "off" pixel."""

    @staticmethod
    def legend_artist(_, __, ___, handlebox):  # noqa: D102
        x0, y0 = calculate_center(handlebox)
        patch = mpatches.RegularPolygon(
            (x0, y0),
            numVertices=6,
            radius=0.7 * handlebox.height,
            orientation=np.deg2rad(30),
            facecolor="black",
            edgecolor="black",
            transform=handlebox.get_transform(),
        )
        handlebox.add_artist(patch)
        return patch


class SquarePixelHandler:
    """Legend handler class to plot a square "on" pixel."""

    @staticmethod
    def legend_artist(_, __, ___, handlebox):  # noqa: D102
        x0, y0 = handlebox.xdescent, handlebox.ydescent
        width = height = handlebox.height
        patch = mpatches.Rectangle(
            [x0, y0],
            width,
            height,
            facecolor=(1, 1, 1, 0),
            edgecolor=(0, 0, 0, 1),
            transform=handlebox.get_transform(),
        )
        handlebox.add_artist(patch)
        return patch


class SquareEdgePixelHandler:
    """Legend handler class to plot a square "edge" pixel."""

    @staticmethod
    def legend_artist(_, __, ___, handlebox):  # noqa: D102
        x0, y0 = handlebox.xdescent, handlebox.ydescent
        width = height = handlebox.height
        patch = mpatches.Rectangle(
            [x0, y0],
            width,
            height,
            facecolor=(*mcolors.to_rgb("brown"), 0.5),
            edgecolor=(*mcolors.to_rgb("black"), 1),
            transform=handlebox.get_transform(),
        )
        handlebox.add_artist(patch)
        return patch


class SquareOffPixelHandler:
    """Legend handler class to plot a square "off" pixel."""

    @staticmethod
    def legend_artist(_, __, ___, handlebox):  # noqa: D102
        x0, y0 = handlebox.xdescent, handlebox.ydescent
        width = height = handlebox.height
        patch = mpatches.Rectangle(
            [x0, y0],
            width,
            height,
            facecolor="black",
            edgecolor="black",
            transform=handlebox.get_transform(),
        )
        handlebox.add_artist(patch)
        return patch


class LSTHandler(TelescopeHandler):
    """Legend handler class to plot a representation of an LST in an array layout."""

    def legend_artist(self, _, __, ___, handlebox):  # noqa: D102
        x0, y0 = calculate_center(handlebox, 10 / 3, 2)
        radius = handlebox.height
        patch = mpatches.Circle(
            xy=(x0, y0),
            radius=radius * self.radius_dict[lst] / self.radius_dict[lst],
            facecolor="none",
            edgecolor=self.colors_dict[lst],
            transform=handlebox.get_transform(),
        )
        handlebox.add_artist(patch)
        return patch


class MSTHandler(TelescopeHandler):
    """Legend handler class to plot a representation of an MST in an array layout."""

    def legend_artist(self, _, __, ___, handlebox):  # noqa: D102
        x0, y0 = calculate_center(handlebox, 4, 2)
        radius = handlebox.height
        patch = mpatches.Circle(
            xy=(x0, y0),
            radius=radius * self.radius_dict[mst] / self.radius_dict[lst],
            facecolor=self.colors_dict[mst],
            edgecolor=self.colors_dict[mst],
            transform=handlebox.get_transform(),
        )
        handlebox.add_artist(patch)
        return patch


class SSTHandler(TelescopeHandler):
    """Legend handler class to plot a representation of an SST in an array layout."""

    def legend_artist(self, _, __, ___, handlebox):  # noqa: D102
        x0, y0 = calculate_center(handlebox, 4, 2)
        radius = handlebox.height
        patch = mpatches.Circle(
            xy=(x0, y0),
            radius=radius * self.radius_dict[sst] / self.radius_dict[lst],
            facecolor=self.colors_dict[sst],
            edgecolor=self.colors_dict[sst],
            transform=handlebox.get_transform(),
        )
        handlebox.add_artist(patch)
        return patch


class SCTHandler(TelescopeHandler):
    """Legend handler class to plot a representation of an SCT in an array layout."""

    def legend_artist(self, _, __, ___, handlebox):  # noqa: D102
        x0, y0 = calculate_center(handlebox, 10, 1)
        width = height = handlebox.height
        patch = mpatches.Rectangle(
            [x0, y0],
            width,
            height,
            facecolor=self.colors_dict[sct],
            edgecolor=self.colors_dict[sct],
            transform=handlebox.get_transform(),
        )
        handlebox.add_artist(patch)
        return patch


class HESSHandler(TelescopeHandler):
    """Legend handler class to plot a representation of an HESS in an array layout."""

    def legend_artist(self, _, __, ___, handlebox):
        x0, y0 = calculate_center(handlebox)
        radius = handlebox.height
        patch = mpatches.RegularPolygon(
            (x0, y0),
            numVertices=6,
            radius=0.7 * radius * self.radius_dict[hess] / self.radius_dict[lst],
            orientation=np.deg2rad(30),
            facecolor=self.colors_dict[hess],
            edgecolor=self.colors_dict[hess],
            transform=handlebox.get_transform(),
        )
        handlebox.add_artist(patch)
        return patch


class MAGICHandler(TelescopeHandler):
    """Legend handler class to plot a representation of an MAGIC in an array layout."""

    def legend_artist(self, _, __, ___, handlebox):
        x0, y0 = calculate_center(handlebox)
        radius = handlebox.height
        patch = mpatches.RegularPolygon(
            (x0, y0),
            numVertices=6,
            radius=0.7 * radius * self.radius_dict[magic] / self.radius_dict[lst],
            orientation=np.deg2rad(30),
            facecolor=self.colors_dict[magic],
            edgecolor=self.colors_dict[magic],
            transform=handlebox.get_transform(),
        )
        handlebox.add_artist(patch)
        return patch


class VERITASHandler(TelescopeHandler):
    """Legend handler class to plot a representation of an VERITAS in an array layout."""

    def legend_artist(self, _, __, ___, handlebox):
        x0, y0 = calculate_center(handlebox)
        radius = handlebox.height
        patch = mpatches.RegularPolygon(
            (x0, y0),
            numVertices=6,
            radius=0.7 * radius * self.radius_dict[veritas] / self.radius_dict[lst],
            orientation=np.deg2rad(30),
            facecolor=self.colors_dict[veritas],
            edgecolor=self.colors_dict[veritas],
            transform=handlebox.get_transform(),
        )
        handlebox.add_artist(patch)
        return patch


class MeanRadiusOuterEdgeHandler:
    """Legend handler class to plot a the mean radius outer edge of the dish."""

    @staticmethod
    def legend_artist(_, __, ___, handlebox):  # noqa: D102
        x0, y0 = calculate_center(handlebox, 4, 4)
        radius = handlebox.height
        patch = mpatches.Circle(
            xy=(x0, y0),
            radius=radius,
            facecolor="none",
            edgecolor="darkorange",
            transform=handlebox.get_transform(),
        )
        handlebox.add_artist(patch)
        return patch


all_telescope_objects = {
    "LSTN": LSTObject,
    "LSTS": LSTObject,
    "MSTN": MSTObject,
    "MSTS": MSTObject,
    sct: SCTObject,
    sst: SSTObject,
    hess: HESSObject,
    magic: MAGICObject,
    veritas: VERITASObject,
}
all_telescope_handlers = {
    "LSTN": LSTHandler,
    "LSTS": LSTHandler,
    "MSTN": MSTHandler,
    "MSTS": MSTHandler,
    sct: SCTHandler,
    sst: SSTHandler,
    hess: HESSHandler,
    magic: MAGICHandler,
    veritas: VERITASHandler,
}
legend_handler_map = {}
try:
    for tel_type in all_telescope_objects.keys():
        legend_handler_map[all_telescope_objects[tel_type]] = all_telescope_handlers[tel_type]
except KeyError:
    pass



#!/usr/bin/python3

"""
Reduces and compiles event data from multiple input files into a structured dataset with event info.

Command line arguments
----------------------
prefix (str, required)
    Path prefix for the input files.
output_file (str, required)
    Path to save the output file.
max_files (int, optional, default=100)
    Maximum number of files to process.
print_dataset_information (flag)
    Print information about the datasets in the generated reduced event dataset.

Example
-------
Generate a reduced dataset from input files and save the result.

.. code-block:: console

    simtools-production-extract-mc-event-data \
        --prefix path/to/input_files/ \
        --wildcard 'gamma_*dark*.simtel.zst' \
        --output_file output_file.hdf5 \
        --max_files 50 \
        --print_dataset_information
"""

import logging
from pathlib import Path

import simtools.utils.general as gen
from simtools.configuration import configurator
from simtools.io_operations import io_handler
from simtools.production_configuration.extract_mc_event_data import MCEventExtractor


def _parse(label, description):
    """
    Parse command line arguments.

    Returns
    -------
    dict
        Parsed command-line arguments.
    """
    config = configurator.Configurator(label=label, description=description)

    config.parser.add_argument(
        "--prefix", type=str, required=True, help="Prefix path for input files."
    )
    config.parser.add_argument(
        "--wildcard",
        type=str,
        required=True,
        help="Wildcard for querying the files in the directory (e.g., 'gamma_*dark*.simtel.zst')",
    )
    config.parser.add_argument("--output_file", type=str, required=True, help="Output filename.")
    config.parser.add_argument(
        "--max_files", type=int, default=100, help="Maximum number of files to process."
    )

    config.parser.add_argument(
        "--print_dataset_information",
        action="store_true",
        help="Print information about the datasets in the generated reduced event dataset.",
    )

    return config.initialize(db_config=False)


def main():
    """
    Process event data files and store data in reduced dataset.

    The reduced dataset contains the following information:
        - simulated: List of simulated events.
        - shower_id_triggered: List of triggered shower IDs
            (as in the telescope definition file used for simulations).
        - triggered_energies: List of energies for triggered events.
        - num_triggered_telescopes: Number of triggered telescopes for each event.
        - core_x: X-coordinate of the shower core (ground coordinates).
        - core_y: Y-coordinate of the shower core (ground coordinates).
        - trigger_telescope_list_list: List of lists containing triggered telescope IDs.
        - file_names: List of input file names.
        - shower_sim_azimuth: Simulated azimuth angle of the shower.
        - shower_sim_altitude: Simulated altitude angle of the shower.
        - array_altitudes: List of altitudes for the array.
        - array_azimuths: List of azimuths for the array.
    """
    label = Path(__file__).stem

    args_dict, _ = _parse(
        label=label,
        description=(
            "Process files and store reduced dataset with event information, "
            "array information and triggered telescopes."
        ),
    )

    _logger = logging.getLogger()
    _logger.setLevel(gen.get_log_level_from_user(args_dict["log_level"]))
    _logger.info(f"Loading input files with prefix: {args_dict['prefix']}")

    input_path = Path(args_dict["prefix"])
    files = list(input_path.glob(args_dict["wildcard"]))
    if not files:
        _logger.warning("No matching input files found.")
        return

    output_path = io_handler.IOHandler().get_output_directory(label)
    output_filepath = Path(output_path).joinpath(f"{args_dict['output_file']}")

    output_filepath.parent.mkdir(parents=True, exist_ok=True)
    generator = MCEventExtractor(files, output_filepath, args_dict["max_files"])
    generator.process_files()
    _logger.info(f"reduced dataset saved to: {output_filepath}")
    if args_dict["print_dataset_information"]:
        generator.print_dataset_information()


if __name__ == "__main__":
    main()



"""Software version setting."""

# this is adapted from https://github.com/cta-observatory/ctapipe/blob/main/ctapipe/version.py
# which is adapted from https://github.com/astropy/astropy/blob/master/astropy/version.py
# see https://github.com/astropy/astropy/pull/10774 for a discussion on why this needed.

try:
    try:
        from ._dev_version import version
    except ImportError:
        from ._version import version
except Exception:  # pylint: disable=broad-except
    import warnings

    warnings.warn("Could not determine simtools version; this indicates a broken installation.")
    del warnings
    version = "0.0.0"  # pylint: disable=invalid-name

__version__ = version



"""Single photon electron spectral analysis."""

import logging
import re
import subprocess
import tempfile
from io import BytesIO
from pathlib import Path

from astropy.table import Table

import simtools.data_model.model_data_writer as writer
from simtools.constants import MODEL_PARAMETER_SCHEMA_URL, SCHEMA_PATH
from simtools.data_model import validate_data
from simtools.data_model.metadata_collector import MetadataCollector
from simtools.io_operations import io_handler


class SinglePhotonElectronSpectrum:
    """
    Single photon electron spectral analysis.

    Parameters
    ----------
    args_dict: dict
        Dictionary with input arguments.
    """

    prompt_column = "frequency (prompt)"
    prompt_plus_afterpulse_column = "frequency (prompt+afterpulsing)"
    afterpulse_column = "frequency (afterpulsing)"

    input_schema = SCHEMA_PATH / "input" / "single_pe_spectrum.schema.yml"

    def __init__(self, args_dict):
        """Initialize SinglePhotonElectronSpectrum class."""
        self._logger = logging.getLogger(__name__)
        self._logger.debug("Initialize SinglePhotonElectronSpectrum class.")

        self.args_dict = args_dict
        # default output is of ecsv format
        self.args_dict["output_file"] = str(
            Path(self.args_dict["output_file"]).with_suffix(".ecsv")
        )
        self.io_handler = io_handler.IOHandler()
        self.data = ""  # Single photon electron spectrum data (as string)
        self.args_dict["metadata_product_data_name"] = "single_pe_spectrum"
        self.args_dict["metadata_product_data_url"] = (
            MODEL_PARAMETER_SCHEMA_URL + "/pm_photoelectron_spectrum.schema.yml"
        )
        self.metadata = MetadataCollector(args_dict=self.args_dict)

    def derive_single_pe_spectrum(self):
        """Derive single photon electron spectrum."""
        if self.args_dict.get("use_norm_spe"):
            return self._derive_spectrum_norm_spe()

        raise NotImplementedError(
            "Derivation of single photon electron spectrum using a simtool is not yet implemented."
        )

    def write_single_pe_spectrum(self):
        """
        Write single photon electron spectrum plus metadata to disk.

        Includes writing in simtel and simtools (ecsv) formats.

        """
        simtel_file = self.io_handler.get_output_directory() / Path(
            self.args_dict["output_file"]
        ).with_suffix(".dat")
        self._logger.debug(f"norm_spe output file: {simtel_file}")
        with open(simtel_file, "w", encoding="utf-8") as simtel:
            simtel.write(self.data)

        cleaned_data = re.sub(r"%%%.+", "", self.data)  # remove norm_spe row metadata
        table = Table.read(
            BytesIO(cleaned_data.encode("utf-8")),
            format="ascii.no_header",
            comment="#",
            delimiter="\t",
        )
        table.rename_columns(
            ["col1", "col2", "col3"],
            ["amplitude", self.prompt_column, self.prompt_plus_afterpulse_column],
        )

        writer.ModelDataWriter.dump(
            args_dict=self.args_dict,
            metadata=self.metadata,
            product_data=table,
            validate_schema_file=None,
        )

    def _derive_spectrum_norm_spe(self):
        """
        Derive single photon electron spectrum using sim_telarray tool 'norm_spe'.

        Returns
        -------
        int
            Return code of the executed command

        Raises
        ------
        subprocess.CalledProcessError
            If the command execution fails.
        """
        tmp_input_file = self._get_input_data(
            input_file=self.args_dict["input_spectrum"],
            frequency_column=self.prompt_column,
        )
        tmp_ap_file = self._get_input_data(
            input_file=self.args_dict.get("afterpulse_spectrum"),
            frequency_column=self.afterpulse_column,
        )

        command = [
            f"{self.args_dict['simtel_path']}/sim_telarray/bin/norm_spe",
            "-r",
            f"{self.args_dict['step_size']},{self.args_dict['max_amplitude']}",
            tmp_input_file.name,
        ]
        if tmp_ap_file:
            command.insert(1, "-a")
            command.insert(2, f"{tmp_ap_file.name}")

        self._logger.debug(f"Running norm_spe command: {' '.join(command)}")
        try:
            result = subprocess.run(command, capture_output=True, text=True, check=True)
        except subprocess.CalledProcessError as exc:
            self._logger.error(f"Error running norm_spe: {exc}")
            self._logger.error(f"stderr: {exc.stderr}")
            raise exc
        finally:
            for tmp_file in [tmp_input_file, tmp_ap_file]:
                try:
                    Path(tmp_file.name).unlink()
                except (AttributeError, FileNotFoundError):
                    pass

        self.data = result.stdout
        return result.returncode

    def _get_input_data(self, input_file, frequency_column):
        """
        Return input data for norm_spe command.

        Input data need to be space separated values of the amplitude spectrum.
        """
        input_data = ""
        if not input_file:
            return None
        input_file = Path(input_file)

        if input_file.suffix == ".ecsv":
            data_validator = validate_data.DataValidator(
                schema_file=self.input_schema, data_file=input_file
            )
            table = data_validator.validate_and_transform()
            input_data = "\n".join(f"{row['amplitude']} {row[frequency_column]}" for row in table)
        else:
            with open(input_file, encoding="utf-8") as f:
                input_data = (
                    f.read().replace(",", " ")
                    if frequency_column == self.prompt_column
                    else f.read()
                )

        with tempfile.NamedTemporaryFile(delete=False, mode="w", encoding="utf-8") as tmpfile:
            tmpfile.write(input_data)
        return tmpfile



#!/usr/bin/python3

r"""
Calculates array or single-telescope trigger rates.

The applications reads from a simtel_array output file, a list of
simtel_array output files ou from a file containing a list of simtel_array files.


Command line arguments
----------------------
simtel_file_names (str or list):
    Path to the simtel_array file or a list of simtel_array output files.
    Files can be generated in `simulate_prod` using the ``--save_file_lists`` option.
save_tables (bool):
    If true, save the tables with the energy-dependent trigger rate to a ecsv file.
area_from_distribution (bool):
    If true, the area thrown (the area in which the simulated events are distributed)
    in the trigger rate calculation is estimated based on the event distribution.
    The expected shape of the distribution of events as function of the core distance is triangular
    up to the maximum distance. The weighted mean radius of the triangular distribution is 2/3 times
    the upper edge. Therefore, when using the ``area_from_distribution`` flag, the mean distance
    times 3/2, returns just the position of the upper edge in the triangle distribution with little
    impact of the binning and little dependence on the scatter area defined in the simulation.
    This is special useful when calculating trigger rate for individual telescopes.
    If false, the area thrown is estimated based on the maximum distance as given in
    the simulation configuration.

Example
-------
Calculate trigger rate from simtel_array file

.. code-block:: console

    simtools-calculate-trigger-rate --simtel_file_names tests/resources/ \\
    run201_proton_za20deg_azm0deg_North_test_layout_test-prod.simtel.zst

Expected final print-out message:

.. code-block:: console

    System trigger rate (Hz): 9.0064e+03 pm 9.0087e+03 Hz

"""

import logging
from pathlib import Path

import simtools.utils.general as gen
from simtools.configuration import configurator
from simtools.io_operations import io_handler
from simtools.simtel.simtel_io_histograms import SimtelIOHistograms


def _parse(label, description):
    """
    Parse command line configuration.

    Parameters
    ----------
    label: str
        Label describing the application.
    description: str
        Description of the application.

    Returns
    -------
    CommandLineParser
        Command line parser object

    """
    config = configurator.Configurator(label=label, description=description)

    config.parser.add_argument(
        "--simtel_file_names",
        help="Name of the simtel_array output files to be calculate the trigger rate from or the "
        "text file containing the list of simtel_array output files.",
        nargs="+",
        required=True,
        type=str,
    )

    config.parser.add_argument(
        "--save_tables",
        help="Save trigger rates per energy bin into ECSV files.",
        action="store_true",
    )

    config.parser.add_argument(
        "--area_from_distribution",
        help="Calculate trigger rates using the event distribution.",
        action="store_true",
    )

    config.parser.add_argument(
        "--stack_files",
        help="Stacks all histograms.",
        action="store_true",
    )

    config_parser, _ = config.initialize(
        db_config=False,
        paths=True,
        simulation_configuration={"corsika_configuration": ["energy_range", "view_cone"]},
    )

    return config_parser


def _get_simulation_parameters(config_parser):
    """
    Get energy range and view cone in the correct form to use in the simtel classes.

    Parameters
    ----------
    CommandLineParser:
        Command line parser object as defined by the _parse function.

    Returns
    -------
    list:
        The energy range used in the simulation.
    list:
        The view cone used in the simulation.

    """

    def convert(param, unit):
        return [param[0].to(unit).value, param[1].to(unit).value] if param else None

    return convert(config_parser.get("energy_range"), "TeV"), convert(
        config_parser.get("view_cone"), "deg"
    )


def main():  # noqa: D103
    label = Path(__file__).stem
    description = (
        "Calculates the simulated and triggered event rate based on simtel_array output files."
    )
    config_parser = _parse(label, description)

    logger = logging.getLogger()
    logger.setLevel(gen.get_log_level_from_user(config_parser["log_level"]))

    simtel_array_files = gen.get_list_of_files_from_command_line(
        config_parser["simtel_file_names"], [".zst", ".simtel", ".hdata"]
    )
    energy_range, view_cone = _get_simulation_parameters(config_parser)

    histograms = SimtelIOHistograms(
        simtel_array_files,
        area_from_distribution=config_parser["area_from_distribution"],
        energy_range=energy_range,
        view_cone=view_cone,
    )

    logger.info("Calculating simulated and triggered event rate")
    (
        sim_event_rates,
        triggered_event_rates,
        triggered_event_rate_uncertainties,
        trigger_rate_in_tables,
    ) = histograms.calculate_trigger_rates(
        print_info=True, stack_files=config_parser["stack_files"]
    )

    # Print out results
    for i_hist, _ in enumerate(sim_event_rates):
        print(f"\nFile {histograms.histogram_files[i_hist]}\n")
        print(
            f"System trigger rate (Hz): {triggered_event_rates[i_hist].value:.4e} \u00b1 "
            f"{triggered_event_rate_uncertainties[i_hist].value:.4e} Hz"
        )
    if config_parser["save_tables"]:
        io_handler_instance = io_handler.IOHandler()
        output_path = io_handler_instance.get_output_directory(label, sub_dir="application-plots")
        for i_table, table in enumerate(trigger_rate_in_tables):
            output_file = (
                str(output_path.joinpath(Path(simtel_array_files[i_table]).stem)) + ".ecsv"
            )
            logger.info(f"Writing table {i_table + 1} to {output_file}")
            table.write(output_file, overwrite=True)


if __name__ == "__main__":
    main()



"""Simulator class for managing simulations of showers and array of telescopes."""

import logging
import re
import shutil
import tarfile
from collections import defaultdict
from pathlib import Path

import numpy as np

import simtools.utils.general as gen
from simtools.corsika.corsika_config import CorsikaConfig
from simtools.io_operations import io_handler
from simtools.job_execution.job_manager import JobManager
from simtools.model.array_model import ArrayModel
from simtools.runners.corsika_runner import CorsikaRunner
from simtools.runners.corsika_simtel_runner import CorsikaSimtelRunner
from simtools.simtel.simulator_array import SimulatorArray

__all__ = [
    "InvalidRunsToSimulateError",
    "Simulator",
]


class InvalidRunsToSimulateError(Exception):
    """Exception for invalid runs to simulate."""


class Simulator:
    """
    Simulator is managing the simulation of showers and of the array of telescopes.

    It interfaces with simulation software packages (e.g., CORSIKA or sim_telarray).

    The configuration is set as a dict corresponding to the command line configuration groups
    (especially simulation_software, simulation_model, simulation_parameters).

    Parameters
    ----------
    args_dict : dict
        Configuration dictionary
        (includes simulation_software, simulation_model, simulation_parameters groups).
    label: str
        Instance label.
    extra_commands: str or list of str
        Extra commands to be added to the run script before the run command.
    mongo_db_config: dict
        MongoDB configuration.
    """

    def __init__(
        self,
        args_dict,
        label=None,
        extra_commands=None,
        mongo_db_config=None,
    ):
        """Initialize Simulator class."""
        self._logger = logging.getLogger(__name__)
        self.args_dict = args_dict

        self.simulation_software = self.args_dict["simulation_software"]
        self._logger.debug(f"Init Simulator {self.simulation_software}")
        self.label = label

        self.io_handler = io_handler.IOHandler()

        self.runs = self._initialize_run_list()
        self._results = defaultdict(list)
        self._test = self.args_dict.get("test", False)
        self.submit_engine = self.args_dict.get("submit_engine", "local")
        self._submit_options = self.args_dict.get("submit_options", None)
        self._extra_commands = extra_commands

        self.array_model = self._initialize_array_model(mongo_db_config)
        self._simulation_runner = self._initialize_simulation_runner(mongo_db_config)

    @property
    def simulation_software(self):
        """The attribute simulation_software."""
        return self._simulation_software

    @simulation_software.setter
    def simulation_software(self, simulation_software):
        """
        Set and test simulation_software type.

        Parameters
        ----------
        simulation_software: choices: [simtel, corsika, corsika_simtel]
            implemented are sim_telarray and CORSIKA or corsika_simtel
            (running CORSIKA and piping it directly to sim_telarray)

        Raises
        ------
        gen.InvalidConfigDataError

        """
        if simulation_software not in ["simtel", "corsika", "corsika_simtel"]:
            self._logger.error(f"Invalid simulation software: {simulation_software}")
            raise gen.InvalidConfigDataError
        self._simulation_software = simulation_software.lower()

    def _initialize_array_model(self, mongo_db_config):
        """
        Initialize array simulation model.

        Parameters
        ----------
        mongo_db_config: dict
            Database configuration.

        Returns
        -------
        ArrayModel
            ArrayModel object.
        """
        return ArrayModel(
            label=self.label,
            site=self.args_dict.get("site"),
            layout_name=self.args_dict.get("array_layout_name"),
            mongo_db_config=mongo_db_config,
            model_version=self.args_dict.get("model_version", None),
        )

    def _initialize_run_list(self):
        """
        Initialize run list using the configuration values 'run_number_start' and 'number_of_runs'.

        Returns
        -------
        list
            List of run numbers.

        Raises
        ------
        KeyError
            If 'run_number_start' or 'number_of_runs' are not found in the configuration.
        """
        try:
            return self._validate_run_list_and_range(
                run_list=None,
                run_range=[
                    self.args_dict["run_number_start"],
                    self.args_dict["run_number_start"] + self.args_dict["number_of_runs"],
                ],
            )
        except KeyError as exc:
            self._logger.error(
                "Error in initializing run list (missing 'run_number_start' or 'number_of_runs')"
            )
            raise exc

    def _validate_run_list_and_range(self, run_list, run_range):
        """
        Prepare list of run numbers from a list or from a range.

        If both arguments are given, they will be merged into a single list.

        Attributes
        ----------
        run_list: list
            list of runs (integers)
        run_range:list
            min and max of range of runs to be simulated (two list entries)

        Returns
        -------
        list
            list of unique run numbers (integers)

        """
        if run_list is None and run_range is None:
            self._logger.debug("Nothing to validate - run_list and run_range not given.")
            return None

        validated_runs = []
        if run_list is not None:
            if not isinstance(run_list, list):
                run_list = [run_list]
            if not all(isinstance(r, int) for r in run_list):
                msg = "run_list must contain only integers."
                self._logger.error(msg)
                raise InvalidRunsToSimulateError
            validated_runs = list(run_list)

        if run_range is not None:
            if not all(isinstance(r, int) for r in run_range) or len(run_range) != 2:
                msg = "run_range must contain two integers only."
                self._logger.error(msg)
                raise InvalidRunsToSimulateError

            run_range = np.arange(run_range[0], run_range[1])
            self._logger.debug(f"run_range: {run_range}")
            validated_runs.extend(list(run_range))

        validated_runs_unique = sorted(set(validated_runs))
        self._logger.info(f"run_list: {validated_runs_unique}")
        return list(validated_runs_unique)

    def _initialize_simulation_runner(self, db_config):
        """
        Initialize corsika configuration and simulation runners.

        Parameters
        ----------
        db_config: dict
            Database configuration.

        Returns
        -------
        CorsikaRunner or SimulatorArray or CorsikaSimtelRunner
            Simulation runner object.
        """
        corsika_config = CorsikaConfig(
            array_model=self.array_model,
            label=self.label,
            args_dict=self.args_dict,
            db_config=db_config,
        )

        runner_class = {
            "corsika": CorsikaRunner,
            "simtel": SimulatorArray,
            "corsika_simtel": CorsikaSimtelRunner,
        }.get(self.simulation_software)

        runner_args = {
            "label": self.label,
            "corsika_config": corsika_config,
            "simtel_path": self.args_dict.get("simtel_path"),
            "use_multipipe": runner_class is CorsikaSimtelRunner,
        }

        if runner_class is not SimulatorArray:
            runner_args["keep_seeds"] = self.args_dict.get("corsika_test_seeds", False)
        if runner_class is not CorsikaRunner:
            runner_args["sim_telarray_seeds"] = self.args_dict.get("sim_telarray_seeds")

        return runner_class(**runner_args)

    def _fill_results_without_run(self, input_file_list):
        """
        Fill results dict without calling submit (e.g., for testing).

        Parameters
        ----------
        input_file_list: str or list of str
            Single file or list of files of shower simulations.
        """
        input_file_list = self._enforce_list_type(input_file_list)

        for file in input_file_list:
            run = self._guess_run_from_file(file)
            self._fill_results(file, run)
            if run not in self.runs:
                self.runs.append(run)

    def simulate(self, input_file_list=None):
        """
        Submit a run script as a job.

        Parameters
        ----------
        input_file_list: str or list of str
            Single file or list of files of shower simulations.
        """
        self._logger.info(f"Submission command: {self.submit_engine}")

        runs_and_files_to_submit = self._get_runs_and_files_to_submit(
            input_file_list=input_file_list
        )
        self._logger.info(
            f"Starting submission for {len(runs_and_files_to_submit)} "
            f"run{'s' if len(runs_and_files_to_submit) > 1 else ''}"
        )

        for run_number, input_file in runs_and_files_to_submit.items():
            run_script = self._simulation_runner.prepare_run_script(
                run_number=run_number, input_file=input_file, extra_commands=self._extra_commands
            )

            job_manager = JobManager(
                submit_engine=self.submit_engine,
                submit_options=self._submit_options,
                test=self._test,
            )
            job_manager.submit(
                run_script=run_script,
                run_out_file=self._simulation_runner.get_file_name(
                    file_type="sub_log", run_number=run_number
                ),
                log_file=self._simulation_runner.get_file_name(
                    file_type=("log"), run_number=run_number
                ),
            )

            self._fill_results(input_file, run_number)

    def _get_runs_and_files_to_submit(self, input_file_list=None):
        """
        Return a dictionary with run numbers and simulation files.

        The latter are expected to be given for the simtel simulator.

        Parameters
        ----------
        input_file_list: str or list of str
            Single file or list of files of shower simulations.

        Returns
        -------
        runs_and_files: dict
            dictionary with run number as key and (if available) simulation
            file name as value

        Raises
        ------
        ValueError
            If no runs are to be submitted.

        """
        _runs_and_files = {}
        self._logger.debug(f"Getting runs and files to submit ({input_file_list})")

        if self.simulation_software == "simtel":
            input_file_list = self._enforce_list_type(input_file_list)
            _runs_and_files = {self._guess_run_from_file(file): file for file in input_file_list}
        elif self.simulation_software in ["corsika", "corsika_simtel"]:
            _runs_and_files = {run: None for run in self._get_runs_to_simulate()}
        if len(_runs_and_files) == 0:
            raise ValueError("No runs to submit.")
        return _runs_and_files

    @staticmethod
    def _enforce_list_type(input_file_list):
        """
        Enforce the input list to be a list.

        Parameters
        ----------
        input_file_list: str or list of str
            Single file or list of files of shower simulations.

        Returns
        -------
        list
            List of input files.
        """
        if not input_file_list:
            return []
        return input_file_list if isinstance(input_file_list, list) else [input_file_list]

    def _guess_run_from_file(self, file):
        """
        Extract the run number from the given file name.

        Input file names can follow any pattern with the
        string 'run' followed by the run number.

        Parameters
        ----------
        file: Path
            Simulation file name

        Returns
        -------
        int
            The extracted run number. If extraction fails, returns 1 and logs a warning.
        """
        file_name = str(Path(file).name)

        try:
            run_str = re.search(r"run\d*", file_name).group()
            return int(run_str[3:])
        except (ValueError, AttributeError):
            self._logger.warning(f"Run number could not be guessed from {file_name} using run = 1")
            return 1

    def _fill_results(self, file, run_number):
        """
        Fill the results dict with input, output, hist, and log files.

        Parameters
        ----------
        file: str
            input file name
        run_number: int
            run number

        """
        keys = ["output", "sub_out", "log", "input", "hist", "corsika_log"]
        defaults = {key: None for key in keys}
        results = {key: defaults[key] for key in keys}
        results["output"] = str(
            self._simulation_runner.get_file_name(file_type="output", run_number=run_number)
        )
        results["sub_out"] = str(
            self._simulation_runner.get_file_name(
                file_type="sub_log", mode="out", run_number=run_number
            )
        )

        if "simtel" in self.simulation_software:
            results["log"] = str(
                self._simulation_runner.get_file_name(
                    file_type="log", simulation_software="simtel", run_number=run_number
                )
            )
            results["input"] = str(file)
            results["hist"] = str(
                self._simulation_runner.get_file_name(
                    file_type="histogram", simulation_software="simtel", run_number=run_number
                )
            )

        if "corsika" in self.simulation_software:
            results["corsika_log"] = str(
                self._simulation_runner.get_file_name(
                    file_type="corsika_log", simulation_software="corsika", run_number=run_number
                )
            )

        for key in keys:
            self._results[key].append(results[key])

    def get_file_list(self, file_type="output"):
        """
        Get list of files generated by simulations.

        Options are "input", "output", "hist", "log", "corsika_log".
        Not all file types are available for all simulation types.
        Returns an empty list for an unknown file type.

        Parameters
        ----------
        file_type : str
            File type to be listed.

        Returns
        -------
        list
            List with the full path of all output files.

        """
        self._logger.info(f"Getting list of {file_type} files")
        return self._results[file_type]

    def print_list_of_files(self, file_type="output"):
        """
        Print list of output files generated by simulations.

        Options are "input", "output", "hist", "log".

        Parameters
        ----------
        file_type : str
            File type to be listed.

        """
        self._logger.info(f"Printing list of {file_type} files")
        for file in self._results[file_type]:
            print(file)

    def _make_resources_report(self, input_file_list):
        """
        Prepare a simple report on computing wall clock time used in the simulations.

        Parameters
        ----------
        input_file_list: str or list of str
            Single file or list of files of shower simulations.

        Returns
        -------
        dict
           Dictionary with reports on computing resources

        """
        if len(self._results["sub_out"]) == 0:
            if input_file_list is None:
                return {"Wall time/run [sec]": np.nan}
            self._fill_results_without_run(input_file_list)

        runtime = []

        _resources = {}
        for run in self.runs:
            _resources = self._simulation_runner.get_resources(run_number=run)
            if _resources.get("runtime"):
                runtime.append(_resources["runtime"])

        mean_runtime = np.mean(runtime)

        resource_summary = {}
        resource_summary["Wall time/run [sec]"] = mean_runtime
        if "n_events" in _resources and _resources["n_events"] > 0:
            resource_summary["#events/run"] = _resources["n_events"]
            resource_summary["Wall time/1000 events [sec]"] = (
                mean_runtime * 1000 / _resources["n_events"]
            )

        return resource_summary

    def resources(self, input_file_list=None):
        """
        Print a simple report on computing resources used.

        Includes run time per run only at this point.

        Parameters
        ----------
        input_file_list: str or list of str
            Single file or list of files of shower simulations.

        """
        resources = self._make_resources_report(input_file_list)
        print("-----------------------------")
        print(f"Computing Resources Report - {self.simulation_software} Simulations")
        for key, value in resources.items():
            print(f"{key} = {value:.2f}")
        print("-----------------------------")

    def _get_runs_to_simulate(self, run_list=None, run_range=None):
        """
        Process run_list and run_range and return the validated list of runs.

        Attributes
        ----------
        run_list: list
            list of runs (integers)
        run_range:list
            min and max of range of runs to be simulated (two list entries)

        Returns
        -------
        list
            list of unique run numbers (integers)

        """
        if run_list is None and run_range is None:
            return [] if self.runs is None else self.runs
        return self._validate_run_list_and_range(run_list, run_range)

    def save_file_lists(self):
        """Save files lists for output and log files."""
        for file_type in ["output", "log", "corsika_log", "hist"]:
            file_name = self.io_handler.get_output_directory(label=self.label).joinpath(
                f"{file_type}_files.txt"
            )
            file_list = self.get_file_list(file_type=file_type)
            if all(element is not None for element in file_list) and len(file_list) > 0:
                self._logger.info(f"Saving list of {file_type} files to {file_name}")
                with open(file_name, "w", encoding="utf-8") as f:
                    for line in self.get_file_list(file_type=file_type):
                        f.write(f"{line}\n")
            else:
                self._logger.debug(f"No files to save for {file_type} files.")

    def pack_for_register(self, directory_for_grid_upload=None):
        """
        Pack simulation output files for registering on the grid.

        Parameters
        ----------
        directory_for_grid_upload: str
            Directory for the tarball with output files.

        """
        self._logger.info(
            f"Packing the output files for registering on the grid ({directory_for_grid_upload})"
        )
        output_files = self.get_file_list(file_type="output")
        log_files = self.get_file_list(file_type="log")
        corsika_log_files = self.get_file_list(file_type="corsika_log")
        histogram_files = self.get_file_list(file_type="hist")
        tar_file_name = Path(log_files[0]).name.replace("log.gz", "log_hist.tar.gz")
        directory_for_grid_upload = (
            Path(directory_for_grid_upload)
            if directory_for_grid_upload
            else self.io_handler.get_output_directory(label=self.label).joinpath(
                "directory_for_grid_upload"
            )
        )
        directory_for_grid_upload.mkdir(parents=True, exist_ok=True)

        tar_file_name = directory_for_grid_upload.joinpath(tar_file_name)

        with tarfile.open(tar_file_name, "w:gz") as tar:
            files_to_tar = (
                (log_files[:1] if log_files else [])
                + (histogram_files[:1] if histogram_files else [])
                + (corsika_log_files[:1] if corsika_log_files else [])
            )
            for file_to_tar in files_to_tar:
                tar.add(file_to_tar, arcname=Path(file_to_tar).name)

        for file_to_move in [*output_files]:
            source_file = Path(file_to_move)
            destination_file = directory_for_grid_upload / source_file.name
            if destination_file.exists():
                self._logger.warning(f"Overwriting existing file: {destination_file}")
            # Note that this will overwrite previous files which exist in the directory
            # It should be fine for normal production since each run is on a separate node
            # so no files are expected there.
            shutil.move(source_file, destination_file)
        self._logger.info(f"Output files for the grid placed in {directory_for_grid_upload!s}")



"""Read sim_telarray events from file."""

import logging
import math
from copy import copy

import astropy.units as u
import numpy as np
from eventio.simtel import SimTelFile

__all__ = ["InconsistentInputFileError", "SimtelIOEvents"]


class InconsistentInputFileError(Exception):
    """Exception for inconsistent input file."""


class SimtelIOEvents:
    """
    Read sim_telarray events from file.

    sim_telarray files are read with eventio package.

    Parameters
    ----------
    input_files: list
        List of sim_telarray output files (str of Path).
    """

    def __init__(self, input_files=None):
        """Initialize SimtelIOEvents."""
        self._logger = logging.getLogger(__name__)
        self.load_input_files(input_files)
        if self.number_of_files > 0:
            self.load_header_and_summary()

    def load_input_files(self, files=None):
        """
        Store list of input files into input_files attribute.

        Parameters
        ----------
        files: list
            List of sim_telarray files (str or Path).
        """
        if not hasattr(self, "input_files"):
            self.input_files = []

        if files is None:
            msg = "No input file was given"
            self._logger.debug(msg)
            return

        if not isinstance(files, list):
            files = [files]

        for file in files:
            self.input_files.append(file)

    @property
    def number_of_files(self):
        """Return number of files loaded.

        Returns
        -------
        int
            Number of files loaded.
        """
        return len(self.input_files) if hasattr(self, "input_files") else 0

    def load_header_and_summary(self):
        """
        Read MC header from sim_telarray files and store it into _mc_header.

        Also fills summary_events with energy and core radius of triggered events.
        """
        self._number_of_files = len(self.input_files)
        keys_to_grab = [
            "obsheight",
            "n_showers",
            "n_use",
            "core_range",
            "diffuse",
            "viewcone",
            "E_range",
            "spectral_index",
            "B_total",
        ]
        self._mc_header = {}

        def _are_headers_consistent(header0, header1):
            comparison = {}
            for k in keys_to_grab:
                value = header0[k] == header1[k]
                comparison[k] = value if isinstance(value, bool) else all(value)

            return all(comparison)

        is_first_file = True
        number_of_triggered_events = 0
        summary_energy, summary_rcore = [], []
        for file in self.input_files:
            with SimTelFile(file) as f:
                for event in f:
                    en = event["mc_shower"]["energy"]
                    rc = math.sqrt(
                        math.pow(event["mc_event"]["xcore"], 2)
                        + math.pow(event["mc_event"]["ycore"], 2)
                    )

                    summary_energy.append(en)
                    summary_rcore.append(rc)
                    number_of_triggered_events += 1

                if is_first_file:
                    # First file - grabbing parameters
                    self._mc_header.update({k: copy(f.mc_run_headers[0][k]) for k in keys_to_grab})
                else:
                    # Remaining files - Checking whether the parameters are consistent
                    if not _are_headers_consistent(self._mc_header, f.mc_run_headers[0]):
                        msg = "MC header pamameters from different files are inconsistent"
                        self._logger.error(msg)
                        raise InconsistentInputFileError(msg)

                is_first_file = False

        self.summary_events = {
            "energy": np.array(summary_energy),
            "r_core": np.array(summary_rcore),
        }

        # Calculating number of events
        self._mc_header["n_events"] = (
            self._mc_header["n_use"] * self._mc_header["n_showers"] * self._number_of_files
        )
        self._mc_header["n_triggered"] = number_of_triggered_events

    @u.quantity_input(core_max=u.m)
    def count_triggered_events(self, energy_range=None, core_max=None):
        """
        Count number of triggered events within a certain energy range and core radius.

        Parameters
        ----------
        energy_range: Tuple with len 2
            Max and min energy of energy range, e.g. energy_range=(100 * u.GeV, 10 * u.TeV).
        core_max: astropy.Quantity distance
            Maximum core radius for selecting showers, e.g. core_max=1000 * u.m.

        Returns
        -------
        int
            Number of triggered events.
        """
        energy_range = self._validate_energy_range(energy_range)
        core_max = self._validate_core_max(core_max)

        is_in_energy_range = [
            energy_range[0] < e < energy_range[1] for e in self.summary_events["energy"]
        ]
        is_in_core_range = [r < core_max for r in self.summary_events["r_core"]]
        return np.sum(np.array(is_in_energy_range) * np.array(is_in_core_range))

    @u.quantity_input(core_max=u.m)
    def select_events(self, energy_range=None, core_max=None):
        """
        Select sim_telarray events within a certain energy range and core radius.

        Parameters
        ----------
        energy_range: Tuple len 2
            Max and min energy of energy range, e.g. energy_range=(100 * u.GeV, 10 * u.TeV).
        core_max: astropy.Quantity distance
            Maximum core radius for selecting showers, e.g. core_max=1000 * u.m.

        Returns
        -------
        list
            List of events.
        """
        energy_range = self._validate_energy_range(energy_range)
        core_max = self._validate_core_max(core_max)

        selected_events = []
        for file in self.input_files:
            with SimTelFile(file) as f:
                for event in f:
                    energy = event["mc_shower"]["energy"]
                    if energy < energy_range[0] or energy > energy_range[1]:
                        continue

                    x_core = event["mc_event"]["xcore"]
                    y_core = event["mc_event"]["ycore"]
                    r_core = math.sqrt(math.pow(x_core, 2) + math.pow(y_core, 2))
                    if r_core > core_max:
                        continue

                    selected_events.append(event)
        return selected_events

    @u.quantity_input(core_max=u.m)
    def count_simulated_events(self, energy_range=None, core_max=None):
        """
        Determine number of simulated events within a certain energy range and  core radius.

        Use the simulated power law.
        This calculation assumes the simulated spectrum is given by a single power law.

        Parameters
        ----------
        energy_range: Tuple len 2
            Max and min energy of energy range, e.g. energy_range=(100 * u.GeV, 10 * u.TeV).
        core_max: astropy.Quantity distance
            Maximum core radius for selecting showers, e.g. core_max=1000 * u.m.

        Returns
        -------
        int
            Number of simulated events.
        """
        energy_range = self._validate_energy_range(energy_range)
        core_max = self._validate_core_max(core_max)

        # energy factor
        def integral(erange):
            power = self._mc_header["spectral_index"] + 1
            return math.pow(erange[0], power) - math.pow(erange[1], power)

        energy_factor = integral(energy_range) / integral(self._mc_header["E_range"])

        # core factor
        core_factor = math.pow(core_max, 2) / math.pow(self._mc_header["core_range"][1], 2)

        return self._mc_header["n_events"] * energy_factor * core_factor

    def _validate_energy_range(self, energy_range):
        """
        Return the default energy range from mc_header in case energy_range=None.

        Check units, convert it to TeV and return it in the right format, otherwise.
        """
        if energy_range is None:
            return self._mc_header["E_range"]

        if not isinstance(energy_range[0], u.Quantity) or not isinstance(
            energy_range[1], u.Quantity
        ):
            msg = "energy_range must be given as u.Quantity in units of energy"
            self._logger.error(msg)
            raise TypeError(msg)

        try:
            return (energy_range[0].to(u.TeV).value, energy_range[1].to(u.TeV).value)
        except u.core.UnitConversionError as e:
            msg = "energy_range must be in units of energy"
            self._logger.error(msg)
            raise TypeError(msg) from e

    def _validate_core_max(self, core_max):
        """
        Return the default core_max from mc_header in case core_max=None.

        Check units, convert it to m and return it in the right format, otherwise.
        """
        return self._mc_header["core_range"][1] if core_max is None else core_max.to(u.m).value



"""Application configuration."""

import argparse
import logging
import os
import sys
import uuid

import astropy.units as u
from dotenv import load_dotenv

import simtools.configuration.commandline_parser as argparser
from simtools.db.db_handler import jsonschema_db_dict
from simtools.io_operations import io_handler
from simtools.utils import general as gen

__all__ = [
    "Configurator",
    "InvalidConfigurationParameterError",
]


class InvalidConfigurationParameterError(Exception):
    """Exception for Invalid configuration parameter."""


class Configurator:
    """
    Application configuration.

    Allow to set configuration parameters by

    - command line arguments
    - configuration file (yml file)
    - configuration dict when calling the class
    - environmental variables

    Assigns unique ACTIVITY_ID to this configuration (uuid).

    Configuration parameter names are converted always to lower case.

    Parameters
    ----------
    config: dict
        Configuration parameters as dict.
    label: str
        Class label.
    usage: str
        Application usage description.
    description: str
        Text displayed as description.
    epilog: str
        Text display after all arguments.
    """

    def __init__(self, config=None, label=None, usage=None, description=None, epilog=None):
        """Initialize Configurator."""
        self._logger = logging.getLogger(__name__)
        self._logger.debug("Init Configuration")

        self.config_class_init = config
        self.label = label
        self.config = {}
        self.parser = argparser.CommandLineParser(
            prog=self.label,
            usage=usage,
            description=description,
            epilog=epilog,
            formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        )

    def default_config(self, arg_list=None, add_db_config=False):
        """
        Return dictionary of default configuration.

        Parameters
        ----------
        arg_list: list
            List of arguments.
        add_db_config: bool
            Add DB configuration file.

        Returns
        -------
        dict
            Configuration parameters as dict.
        """
        self.parser.initialize_default_arguments()
        simulation_model = None
        if arg_list and "--site" in arg_list:
            simulation_model = ["site"]
        if arg_list and "--telescope" in arg_list:
            simulation_model = ["site", "telescope"]
        self.parser.initialize_simulation_model_arguments(simulation_model)
        if add_db_config:
            self.parser.initialize_db_config_arguments()

        self._fill_config(arg_list)
        return self.config

    def initialize(
        self,
        require_command_line=True,
        paths=True,
        output=False,
        simulation_model=None,
        simulation_configuration=None,
        db_config=False,
        job_submission=False,
    ):
        """
        Initialize application configuration.

        Configure from command line, configuration file, class config, or environmental variable.

        Priorities in parameter settings.
        1. command line; 2. yaml file; 3. class init; 4. env variables.

        Conflicting configuration settings raise an Exception, with the exception of settings \
        from environmental variables, which are only done when the configuration parameter \
        is None.

        Parameters
        ----------
        require_command_line: bool
            Require at least one command line argument.
        paths: bool
            Add path configuration to list of args.
        output: bool
            Add output file configuration to list of args.
        simulation_model: list
            List of simulation model configuration parameters to add to list of args
        simulation_configuration: dict
            Dict of simulation software configuration parameters to add to list of args.
        db_config: bool
            Add database configuration parameters to list of args.
        job_submission: bool
            Add job submission configuration to list of args.

        Returns
        -------
        dict
            Configuration parameters as dict.
        dict
            Dictionary with DB parameters

        """
        self.parser.initialize_default_arguments(
            paths=paths,
            output=output,
            simulation_model=simulation_model,
            simulation_configuration=simulation_configuration,
            db_config=db_config,
            job_submission=job_submission,
        )

        self._fill_from_command_line(require_command_line=require_command_line)
        self._fill_from_config_file(self.config.get("config"))
        self._fill_from_config_dict(self.config_class_init)
        self._fill_from_environmental_variables()

        if self.config.get("activity_id", None) is None:
            self.config["activity_id"] = str(uuid.uuid4())
        if self.config["label"] is None:
            self.config["label"] = self.label

        self._initialize_io_handler()
        if output:
            self._initialize_output()
        _db_dict = self._get_db_parameters()

        return self.config, _db_dict

    def _fill_from_command_line(self, arg_list=None, require_command_line=True):
        """
        Fill configuration parameters from command line arguments.

        Triggers a print of the help if no command line arguments are given and
        require_command_line is set.

        Parameters
        ----------
        arg_list: list
            List of arguments.
        require_command_line: bool
            Require at least one command line argument.

        """
        if arg_list is None:
            arg_list = sys.argv[1:]

        if require_command_line and len(arg_list) == 0:
            self._logger.debug("No command line arguments given, printing help.")
            arg_list = ["--help"]

        if "--config" in arg_list:
            self._reset_required_arguments()

        self._fill_config(arg_list)

    def _reset_required_arguments(self):
        """
        Reset required parser arguments (i.e., arguments added with "required=True").

        Includes also mutually exclusive groups.
        Access protected attributes of parser (no public method available).

        """
        for group in self.parser._mutually_exclusive_groups:  # pylint: disable=protected-access
            group.required = False
        for action in self.parser._actions:  # pylint: disable=protected-access
            action.required = False

    def _fill_from_config_dict(self, input_dict, overwrite=False):
        """
        Fill configuration parameters from dictionary.

        Enforce that configuration parameter names are lower case.

        Parameters
        ----------
        input_dict: dict
            dictionary with configuration parameters.
        overwrite: bool
            overwrite existing configuration parameters.

        """
        _tmp_config = {}
        try:
            for key, value in input_dict.items():
                if not overwrite:
                    self._check_parameter_configuration_status(key, value)
                _tmp_config[key.lower()] = value
        except AttributeError:
            pass

        self._fill_config(_tmp_config)

    def _check_parameter_configuration_status(self, key, value):
        """
        Check if a parameter is already configured and not still set to the default value.

        Allow configuration with None values.

        Parameters
        ----------
        key, value
           parameter key, value to be checked


        Raises
        ------
        InvalidConfigurationParameterError
           if parameter has already been defined with a different value.
        """
        # parameter not changed or None
        if self.parser.get_default(key) == self.config[key] or self.config[key] is None:
            return

        # parameter already set
        if key in self.config and self.config[key] != value:
            self._logger.error(
                f"Inconsistent configuration parameter ({key}) definition "
                f"({self.config[key]} vs {value})"
            )
            raise InvalidConfigurationParameterError

    def _fill_from_config_file(self, config_file):
        """
        Read and fill configuration parameters from yaml file.

        Parameters
        ----------
        config file: str
            Name of configuration file name

        Raises
        ------
        FileNotFoundError
           if configuration file has not been found.

        """
        try:
            self._logger.debug(f"Reading configuration from {config_file}")
            _config_dict = (
                gen.collect_data_from_file(file_name=config_file) if config_file else None
            )
            # yaml parser adds \n in multiline strings, remove them
            _config_dict = gen.remove_substring_recursively_from_dict(_config_dict, substring="\n")
            # read configuration for first application
            if "CONFIGURATION" in _config_dict.get("CTA_SIMPIPE", {}).get("APPLICATIONS", [{}])[0]:
                self._fill_from_config_dict(
                    input_dict=gen.change_dict_keys_case(
                        _config_dict["CTA_SIMPIPE"]["APPLICATIONS"][0]["CONFIGURATION"],
                    ),
                    overwrite=True,
                )
            else:
                self._fill_from_config_dict(
                    input_dict=gen.change_dict_keys_case(_config_dict), overwrite=True
                )
        # TypeError is raised for config_file=None
        except (TypeError, AttributeError):
            pass
        except FileNotFoundError:
            self._logger.error(f"Configuration file not found: {config_file}")
            raise

    def _fill_from_environmental_variables(self):
        """
        Fill any configuration parameters from environmental variables or from file (e.g., ".env").

        Only parameters which are not already configured are changed (i.e., parameter is None).

        """
        _env_dict = {}
        try:
            load_dotenv(self.config["env_file"])
        except KeyError:
            pass
        for key, value in self.config.items():
            # environmental variables for simtools should always start with SIMTOOLS_
            env_variable_to_read = f"SIMTOOLS_{key.upper()}"
            if value is None:
                env_value = os.environ.get(env_variable_to_read)
                if env_value is not None:
                    env_value = env_value.split("#")[0].strip().replace('"', "").replace("'", "")
                _env_dict[key] = env_value

        self._fill_from_config_dict(_env_dict)

    def _initialize_io_handler(self):
        """Initialize IOHandler with input and output paths."""
        _io_handler = io_handler.IOHandler()
        _io_handler.set_paths(
            output_path=self.config.get("output_path", None),
            use_plain_output_path=self.config.get("use_plain_output_path", False),
            data_path=self.config.get("data_path", None),
            model_path=self.config.get("model_path", None),
        )

    def _initialize_output(self):
        """Initialize default output file names (in case output_file is not configured)."""
        if self.config.get("output_file", None) is None:
            self.config["output_file_from_default"] = True
            prefix = "TEST"
            label = extension = ""
            if not self.config.get("test", False):
                prefix = self.config["activity_id"]
                if self.config.get("label", "") and len(self.config.get("label", "")) > 0:
                    label = f"-{self.config['label']}"
            if len(self.config.get("output_file_format", "")) > 0:
                extension = f".{self.config['output_file_format']}"

            self.config["output_file"] = f"{prefix}{label}{extension}"

    @staticmethod
    def _arglist_from_config(input_var):
        """
        Convert input list of strings as needed by argparse.

        Special cases:
        - lists as arguments (using e.g., nargs="+") are expanded
        - boolean are expected to be handled as action="store_true" or "store_false"
        - None values or zero length values are ignored (this means setting a parameter
          to none or "" is not allowed).


        Ignore values which are None or of zero length.

        Parameters
        ----------
        input_var: dict, list, None
           Dictionary/list of commands to convert to list.

        Returns
        -------
        list
            Dict keys and values as dict.

        """
        if isinstance(input_var, dict):
            _list_args = []
            for key, value in input_var.items():
                if isinstance(value, list):
                    _list_args.append("--" + key)
                    _list_args.extend(map(str, value))
                elif isinstance(value, u.Quantity) or (
                    not isinstance(value, bool) and value is not None and len(str(value)) > 0
                ):
                    _list_args.append("--" + key)
                    _list_args.append(str(value))
                elif value:
                    _list_args.append("--" + key)
            return _list_args

        try:
            return [str(value) for value in list(input_var) if value != "None"]
        except TypeError:
            return []

    @staticmethod
    def _convert_string_none_to_none(input_dict):
        """
        Convert string type 'None' to type None (argparse returns None as str).

        Parameters
        ----------
        input_dict
            Dictionary with values to be converted.

        """
        for key, value in input_dict.items():
            input_dict[key] = None if value == "None" else value

        return input_dict

    def _fill_config(self, input_container):
        """
        Fill configuration dictionary.

        Parameters
        ----------
        input_container
            List or dictionary with configuration updates.
        """
        self.config = self._convert_string_none_to_none(
            vars(
                self.parser.parse_args(
                    self._arglist_from_config(self.config)
                    + self._arglist_from_config(input_container)
                )
            )
        )

    def _get_db_parameters(self):
        """
        Return parameters for DB configuration.

        Returns
        -------
        dict
            Dictionary with DB parameters.
        """
        db_params = jsonschema_db_dict["properties"].keys()
        return {param: self.config.get(param) for param in db_params if param in self.config}



"""Simulation runner for array simulations."""

import logging

from simtools.io_operations import io_handler
from simtools.runners.simtel_runner import InvalidOutputFileError, SimtelRunner
from simtools.utils.general import clear_default_sim_telarray_cfg_directories

__all__ = ["SimulatorArray"]


class SimulatorArray(SimtelRunner):
    """
    SimulatorArray is the interface with sim_telarray to perform array simulations.

    Parameters
    ----------
    corsika_config_data: CorsikaConfig
        CORSIKA configuration.
    simtel_path: str or Path
        Location of source of the sim_telarray/CORSIKA package.
    label: str
        Instance label.
    use_multipipe: bool
        Use multipipe to run CORSIKA and sim_telarray.
    """

    def __init__(
        self,
        corsika_config,
        simtel_path,
        label=None,
        use_multipipe=False,
        sim_telarray_seeds=None,
    ):
        """Initialize SimulatorArray."""
        self._logger = logging.getLogger(__name__)
        self._logger.debug("Init SimulatorArray")
        super().__init__(
            label=label,
            simtel_path=simtel_path,
            corsika_config=corsika_config,
            use_multipipe=use_multipipe,
        )

        self.sim_telarray_seeds = sim_telarray_seeds
        self.corsika_config = corsika_config
        self.io_handler = io_handler.IOHandler()
        self._log_file = None

    def _make_run_command(self, run_number=None, input_file=None):
        """
        Build and return the command to run simtel_array.

        Parameters
        ----------
        input_file: str
            Full path of the input CORSIKA file
        run_number: int (optional)
            run number

        Returns
        -------
        str
            Command to run sim_telarray.
        """
        self._log_file = self.get_file_name(file_type="log", run_number=run_number)
        histogram_file = self.get_file_name(file_type="histogram", run_number=run_number)
        output_file = self.get_file_name(file_type="output", run_number=run_number)

        # Array
        command = str(self._simtel_path.joinpath("sim_telarray/bin/sim_telarray"))
        command += f" -c {self.corsika_config.array_model.get_config_file()}"
        command += f" -I{self.corsika_config.array_model.get_config_directory()}"
        command += super().get_config_option("telescope_theta", self.corsika_config.zenith_angle)
        command += super().get_config_option("telescope_phi", self.corsika_config.azimuth_angle)
        command += super().get_config_option(
            "power_law",
            SimulatorArray.get_power_law_for_sim_telarray_histograms(
                self.corsika_config.primary_particle
            ),
        )
        command += super().get_config_option("histogram_file", histogram_file)
        command += super().get_config_option("output_file", output_file)
        command += super().get_config_option("random_state", "none")
        if self.sim_telarray_seeds:
            command += super().get_config_option("random_seed", self.sim_telarray_seeds)
        command += super().get_config_option("show", "all")
        command += f" {input_file}"
        command += f" > {self._log_file} 2>&1 || exit"

        return clear_default_sim_telarray_cfg_directories(command)

    def _check_run_result(self, run_number=None):
        """
        Check if simtel output file exists.

        Parameters
        ----------
        run_number: int
            Run number.

        Returns
        -------
        bool
            True if simtel output file exists.

        Raises
        ------
        InvalidOutputFileError
            If simtel output file does not exist.
        """
        output_file = self.get_file_name(file_type="output", run_number=run_number)
        if not output_file.exists():
            msg = f"sim_telarray output file {output_file} does not exist."
            self._logger.error(msg)
            raise InvalidOutputFileError(msg)
        self._logger.debug(f"simtel_array output file {output_file} exists.")
        return True

    @staticmethod
    def get_power_law_for_sim_telarray_histograms(primary):
        """
        Get the power law index for sim_telarray.

        Events will be histogrammed in sim_telarray with a weight according to
        the difference between this exponent and the one used for the shower simulations.

        Parameters
        ----------
        primary: str
            Primary particle.

        Returns
        -------
        float
            Power law index.
        """
        power_laws = {
            "gamma": 2.5,
            "electron": 3.3,
        }
        if primary.name in power_laws:
            return power_laws[primary.name]

        return 2.68



#!/usr/bin/python3
r"""
    Submit a model parameter value and corresponding metadata through the command line.

    Input and metadata is validated, and if necessary enriched and converted following
    the model parameter schemas. Model parameter data is written in the simtools-style
    json format, metadata as a yaml file.

    Command line arguments
    ----------------------
    parameter (str)
        model parameter name
    value (str, value)
        input value (number, string, string-type lists)
    instrument (str)
        instrument name.
    site (str)
        site location.
    parameter_version (str)
        Parameter version.
    input_meta (str, optional)
        input meta data file (yml format)

    Example
    -------

    Submit the number of gains for the LSTN-design readout chain:

    .. code-block:: console

        simtools-submit-model-parameter-from-external \
            --parameter num_gains \\
            --value 2 \\
            --instrument LSTN-design \\
            --site North \\
            --parameter_version 0.1.0 \\
            --input_meta num_gains.metadata.yml

"""

import logging
from pathlib import Path

import simtools.data_model.model_data_writer as writer
import simtools.utils.general as gen
from simtools.configuration import configurator


def _parse(label, description):
    """
    Parse command line configuration.

    Parameters
    ----------
    label: str
        Label describing application.
    description: str
        Description of application.

    Returns
    -------
    CommandLineParser
        Command line parser object

    """
    config = configurator.Configurator(label=label, description=description)

    config.parser.add_argument(
        "--parameter", type=str, required=True, help="Parameter for simulation model"
    )
    config.parser.add_argument("--instrument", type=str, required=True, help="Instrument name")
    config.parser.add_argument("--site", type=str, required=True, help="Site location")
    config.parser.add_argument(
        "--parameter_version", type=str, required=True, help="Parameter version"
    )
    config.parser.add_argument(
        "--value",
        type=str,
        required=True,
        help=(
            "Model parameter value. "
            "Can be a single number, a number with a unit, or a list of values with units. "
            'Examples: "--value=5", "--value=\'5 km\'", "--value=\'5 cm, 0.5 deg\'"'
        ),
    )
    config.parser.add_argument(
        "--input_meta",
        help="meta data file(s) associated to input data (wildcards or list of files allowed)",
        type=str,
        nargs="+",
        required=False,
    )
    config.parser.add_argument(
        "--check_parameter_version",
        help="Check if the parameter version exists in the database",
        action="store_true",
    )
    return config.initialize(output=True, db_config=True)


def main():  # noqa: D103
    args_dict, db_config = _parse(
        label=Path(__file__).stem,
        description="Submit and validate a model parameters).",
    )

    logger = logging.getLogger()
    logger.setLevel(gen.get_log_level_from_user(args_dict["log_level"]))

    output_path = (
        Path(args_dict["output_path"]) / args_dict["instrument"] / args_dict["parameter"]
        if args_dict.get("output_path")
        else None
    )

    writer.ModelDataWriter.dump_model_parameter(
        parameter_name=args_dict["parameter"],
        value=args_dict["value"],
        instrument=args_dict["instrument"],
        parameter_version=args_dict["parameter_version"],
        output_file=Path(args_dict["parameter"] + "-" + args_dict["parameter_version"] + ".json"),
        output_path=output_path,
        use_plain_output_path=args_dict.get("use_plain_output_path"),
        metadata_input_dict=args_dict,
        db_config=db_config if args_dict.get("check_parameter_version") else None,
    )


if __name__ == "__main__":
    main()



#!/usr/bin/python3
r"""
    Generate a default simtools metadata file from a json schema.

    Command line arguments
    ----------------------
    schema (str, optional)
        Schema file describing the input data
        (default: simtools/schemas/metadata.metaschema.yml)
    output_file (str, optional)
        Output file name.

    Example
    -------
    .. code-block:: console

        simtools-generate-default-metadata \\
            --schema simtools/schemas/metadata.metaschema.yml \\
            --output_file default_metadata.yml


    """

import json
import logging
from pathlib import Path

import yaml

import simtools.utils.general as gen
from simtools.configuration import configurator
from simtools.data_model import metadata_model
from simtools.io_operations import io_handler


def _parse(label, description):
    """
    Parse command line configuration.

    Parameters
    ----------
    label: str
        Label describing application.
    description: str
        Description of application.

    Returns
    -------
    CommandLineParser
        Command line parser object

    """
    config = configurator.Configurator(label=label, description=description)

    config.parser.add_argument(
        "--schema",
        help="schema file describing input data",
        type=str,
        required=True,
    )
    config.parser.add_argument(
        "--output_file",
        help="output file name (if not given: print to stdout)",
        type=str,
        required=False,
    )

    return config.initialize(output=False, require_command_line=True)


def main():  # noqa: D103
    label = Path(__file__).stem
    args_dict, _ = _parse(
        label, description="Generate a default simtools metadata file from a json schema."
    )

    _logger = logging.getLogger()
    _logger.setLevel(gen.get_log_level_from_user(args_dict["log_level"]))

    default_values = metadata_model.get_default_metadata_dict(args_dict["schema"])

    if args_dict["output_file"] is None:
        print(default_values)
    else:
        _io_handler = io_handler.IOHandler()
        _out_file = _io_handler.get_output_file(args_dict["output_file"])
        _logger.info(f"Writing default values to {_out_file}")
        if args_dict["output_file"].endswith((".yml", ".yaml")):
            with open(_out_file, "w", encoding="utf-8") as file:
                yaml.dump(
                    default_values,
                    file,
                    default_flow_style=False,
                    sort_keys=False,
                )
        if args_dict["output_file"].endswith(".json"):
            with open(_out_file, "w", encoding="utf-8") as file:
                json.dump(
                    default_values,
                    file,
                    indent=4,
                    sort_keys=False,
                )


if __name__ == "__main__":
    main()



"""
Calculate the scaled number of events based on metrics.

Module for scaling events based on statistical error metrics. Contains the `EventScaler` class,
which scales the number of events for both the entire dataset and specific grid points.
Scaling factors are calculated using error metrics and the evaluator's results.
"""

import astropy.units as u
import numpy as np

__all__ = ["EventScaler"]


class EventScaler:
    """
    Scales the number of events based on statistical error metrics.

    Supports scaling both the entire dataset and specific grid points like energy values.
    """

    def __init__(self, evaluator, metrics: dict):
        """
        Initialize the EventScaler with the evaluator and metrics.

        Parameters
        ----------
        evaluator : StatisticalErrorEvaluator
            The evaluator responsible for calculating metrics and handling event data.
        metrics : dict
            Dictionary containing metrics, including target error for effective area.
        """
        self.evaluator = evaluator
        self.metrics = metrics

    def scale_events(self, return_sum: bool = True) -> u.Quantity:
        """
        Calculate the scaled number of events based on statistical error metrics.

        Parameters
        ----------
        return_sum : bool, optional
            If True, returns the sum of scaled events for the entire set of MC events. If False,
            returns the scaled events for each grid point along the energy axis. Default is True.

        Returns
        -------
        u.Quantity
            If 'return_sum' is True, returns the total scaled number of events as a u.Quantity.
            If 'return_sum' is False, returns an array of scaled events along the energy axis as
            a u.Quantity.
        """
        scaling_factor = self._compute_scaling_factor()

        base_events = self._number_of_simulated_events()

        if return_sum:
            return np.sum(base_events * scaling_factor)
        return base_events * scaling_factor

    def _compute_scaling_factor(self) -> float:
        """
        Compute the scaling factor based on the error metrics.

        Returns
        -------
        float
            The scaling factor.
        """
        metric_results = self.evaluator.calculate_metrics()
        uncertainty_effective_area = metric_results.get("uncertainty_effective_area", {})
        current_max_error = uncertainty_effective_area.get("max_error")
        target_max_error = self.metrics.get("uncertainty_effective_area", {}).get("target_error")[
            "value"
        ]

        return (current_max_error / target_max_error) ** 2

    def _number_of_simulated_events(self) -> u.Quantity:
        """
        Fetch the number of simulated events from the evaluator's data.

        Returns
        -------
        u.Quantity
            The number of simulated events.
        """
        return self.evaluator.data.get("simulated_event_histogram")

    def calculate_scaled_events_at_grid_point(
        self,
        grid_point: tuple,
    ) -> u.Quantity:
        """
        Calculate the scaled number of events for a specific energy grid point.

        Parameters
        ----------
        grid_point : tuple
            The grid point specifying energy, azimuth, zenith, NSB, and offset.

        Returns
        -------
        float
            The scaled number of events at the specified grid point (energy).
        """
        energy = grid_point[0]
        bin_edges = self.evaluator.create_bin_edges()
        bin_idx = np.digitize(energy, bin_edges) - 1

        scaling_factor = self._compute_scaling_factor()

        simulated_event_histogram = self.evaluator.data.get("simulated_event_histogram", [])

        if bin_idx < 0 or bin_idx >= len(simulated_event_histogram):
            raise ValueError(f"Energy {energy} is outside the range of the simulated events data.")

        base_events = self._number_of_simulated_events()
        base_event_at_energy = base_events[bin_idx]
        return base_event_at_energy * scaling_factor



"""Generate a reduced dataset from given simulation event list and save the output to file."""

import logging

import numpy as np
import tables
from ctapipe.core import Container, Field
from ctapipe.io import HDF5TableWriter
from eventio import EventIOFile
from eventio.simtel import ArrayEvent, MCEvent, MCRunHeader, MCShower, TriggerInformation

DEFAULT_FILTERS = tables.Filters(complevel=5, complib="zlib", shuffle=True, bitshuffle=False)


class ReducedDatasetContainer(Container):
    """Container for reduced dataset information."""

    simulated = Field(None, "Simulated energy")
    core_x = Field(None, "X-coordinate of the shower core")
    core_y = Field(None, "Y-coordinate of the shower core")
    shower_sim_azimuth = Field(None, "Simulated azimuth angle of the shower")
    shower_sim_altitude = Field(None, "Simulated altitude angle of the shower")
    array_altitudes = Field(None, "Altitudes for the array")
    array_azimuths = Field(None, "Azimuths for the array")


class TriggeredShowerContainer(Container):
    """Container for triggered shower information."""

    shower_id_triggered = Field(None, "Triggered shower ID")
    triggered_energies = Field(None, "Triggered energies")


class FileNamesContainer(Container):
    """Container for file names."""

    file_names = Field(None, "Input file names")


class MCEventExtractor:
    """
    Generate a reduced dataset from given simulation event list and save the output to file.

    Attributes
    ----------
    input_files : list
        List of input file paths to process.
    output_file : str
        Path to the output HDF5 file.
    max_files : int, optional
        Maximum number of files to process.
    """

    def __init__(self, input_files, output_file, max_files=100):
        """
        Initialize the MCEventExtractor with input files, output file, and max file limit.

        Parameters
        ----------
        input_files : list
            List of input file paths to process.
        output_file : str
            Path to the output HDF5 file.
        max_files : int, optional
            Maximum number of files to process.
        """
        self._logger = logging.getLogger(__name__)
        self.input_files = input_files
        self.output_file = output_file
        self.max_files = max_files
        self.shower = None
        self.n_use = None
        self.shower_id_offset = 0

    def process_files(self):
        """Process the input files and store them in an HDF5 file."""
        if not self.input_files:
            self._logger.warning("No input files provided.")
            return

        data_lists = self._initialize_data_lists()
        self.shower_id_offset = 0
        # Process the first file in write mode
        self._logger.info(f"Processing file 1/{self.max_files}: {self.input_files[0]}")
        self._process_file(self.input_files[0], data_lists, str(self.input_files[0]))
        self._write_all_data(data_lists, mode="w")
        self.shower_id_offset = len(data_lists["simulated"])
        self._reset_data_lists(data_lists)

        # Process remaining files in append mode
        for i_file, file in enumerate(self.input_files[1 : self.max_files], start=2):
            self._logger.info(f"Processing file {i_file}/{self.max_files}: {file}")
            self._process_file(file, data_lists, str(file))
            if len(data_lists["simulated"]) >= 1e7:
                self._write_all_data(data_lists, mode="a")
                self.shower_id_offset += len(data_lists["simulated"])
                self._reset_data_lists(data_lists)

        # Final write for any remaining data
        self._write_all_data(data_lists, mode="a")

    def _write_all_data(self, data_lists, mode):
        """Write all data sections at once helper method."""
        self._write_data(data_lists, mode=mode)
        self._write_variable_length_data(data_lists["trigger_telescope_list_list"], mode="a")
        self._write_file_names(data_lists["file_names"], mode="a")

    def _write_file_names(self, file_names, mode="a"):
        """Write file names to HDF5 file."""
        print("file_names", file_names)
        with HDF5TableWriter(
            self.output_file, group_name="data", mode=mode, filters=DEFAULT_FILTERS
        ) as writer:
            file_names_container = FileNamesContainer()
            for file_name in file_names:
                file_names_container.file_names = file_name
                writer.write(table_name="file_names", containers=[file_names_container])

    def _write_variable_length_data(self, trigger_telescope_list_list, mode="a"):
        """Write variable-length array data to HDF5 file."""
        with tables.open_file(self.output_file, mode=mode) as f:
            if "trigger_telescope_list_list" in f.root.data:
                vlarray = f.root.data.trigger_telescope_list_list
            else:
                vlarray = f.create_vlarray(
                    f.root.data,
                    "trigger_telescope_list_list",
                    tables.Int16Atom(),
                    "List of triggered telescope IDs",
                )

            for item in trigger_telescope_list_list:
                vlarray.append(item)

    def _initialize_data_lists(self):
        """Initialize data lists."""
        return {
            "simulated": [],
            "shower_id_triggered": [],
            "triggered_energies": [],
            "core_x": [],
            "core_y": [],
            "trigger_telescope_list_list": [],
            "file_names": [],
            "shower_sim_azimuth": [],
            "shower_sim_altitude": [],
            "array_altitudes": [],
            "array_azimuths": [],
        }

    def _process_file(self, file, data_lists, file_name):
        """Process a single file and update data lists."""
        with EventIOFile(file) as f:
            array_altitude = None
            array_azimuth = None
            for eventio_object in f:
                if isinstance(eventio_object, MCRunHeader):
                    self._process_mc_run_header(eventio_object, data_lists)
                elif isinstance(eventio_object, MCShower):
                    self._process_mc_shower(
                        eventio_object, data_lists, array_altitude, array_azimuth
                    )
                elif isinstance(eventio_object, MCEvent):
                    self._process_mc_event(eventio_object, data_lists)
                elif isinstance(eventio_object, ArrayEvent):
                    self._process_array_event(eventio_object, data_lists)
            data_lists["file_names"].extend([file_name])

    def _process_mc_run_header(self, eventio_object, data_lists):
        """Process MC run header and update data lists."""
        mc_head = eventio_object.parse()
        self.n_use = mc_head["n_use"]  # reuse factor n_use needed to extend the values below
        array_altitude = np.mean(mc_head["alt_range"])
        array_azimuth = np.mean(mc_head["az_range"])
        data_lists["array_altitudes"].extend(self.n_use * [array_altitude])
        data_lists["array_azimuths"].extend(self.n_use * [array_azimuth])

    def _process_mc_shower(self, eventio_object, data_lists, array_altitude, array_azimuth):
        """Process MC shower and update data lists."""
        self.shower = eventio_object.parse()
        data_lists["simulated"].extend(self.n_use * [self.shower["energy"]])
        data_lists["shower_sim_azimuth"].extend(self.n_use * [self.shower["azimuth"]])
        data_lists["shower_sim_altitude"].extend(self.n_use * [self.shower["altitude"]])
        data_lists["array_altitudes"].extend(self.n_use * [array_altitude])
        data_lists["array_azimuths"].extend(self.n_use * [array_azimuth])

    def _process_mc_event(self, eventio_object, data_lists):
        """Process MC event and update data lists."""
        event = eventio_object.parse()
        data_lists["core_x"].append(event["xcore"])
        data_lists["core_y"].append(event["ycore"])

    def _process_array_event(self, eventio_object, data_lists):
        """Process array event and update data lists."""
        for i, obj in enumerate(eventio_object):
            if i == 0 and isinstance(obj, TriggerInformation):
                self._process_trigger_information(obj, data_lists)

    def _process_trigger_information(self, trigger_info, data_lists):
        """Process trigger information and update data lists."""
        trigger_info = trigger_info.parse()
        telescopes = trigger_info["telescopes_with_data"]
        if len(telescopes) > 0:
            data_lists["shower_id_triggered"].append(self.shower["shower"] + self.shower_id_offset)
            data_lists["triggered_energies"].append(self.shower["energy"])
            data_lists["trigger_telescope_list_list"].append(np.array(telescopes, dtype=np.int16))

    def _write_data(self, data_lists, mode="a"):
        """Write data to HDF5 file using HDF5TableWriter."""
        with HDF5TableWriter(
            self.output_file, group_name="data", mode=mode, filters=DEFAULT_FILTERS
        ) as writer:
            # Write reduced dataset container
            reduced_container = ReducedDatasetContainer()
            for i in range(len(data_lists["simulated"])):
                reduced_container.simulated = data_lists["simulated"][i]
                reduced_container.core_x = data_lists["core_x"][i]
                reduced_container.core_y = data_lists["core_y"][i]
                reduced_container.shower_sim_azimuth = data_lists["shower_sim_azimuth"][i]
                reduced_container.shower_sim_altitude = data_lists["shower_sim_altitude"][i]
                reduced_container.array_altitudes = data_lists["array_altitudes"][i]
                reduced_container.array_azimuths = data_lists["array_azimuths"][i]
                writer.write(table_name="reduced_data", containers=[reduced_container])

            # Write triggered shower container
            triggered_container = TriggeredShowerContainer()
            for i in range(len(data_lists["shower_id_triggered"])):
                triggered_container.shower_id_triggered = data_lists["shower_id_triggered"][i]
                triggered_container.triggered_energies = data_lists["triggered_energies"][i]
                writer.write(table_name="triggered_data", containers=[triggered_container])

    def _reset_data_lists(self, data_lists):
        """Reset data lists during batch processing."""
        for key in data_lists:
            data_lists[key] = []

    def print_dataset_information(self):
        """Print information about the datasets in the generated HDF5 file."""
        try:
            with tables.open_file(self.output_file, mode="r") as reader:
                print("Datasets in file:")
                for key in reader.root.data._v_children.keys():  # pylint: disable=protected-access
                    dset = reader.root.data._v_children[key]  # pylint: disable=protected-access
                    print(f"- {key}: shape={dset.shape}, dtype={dset.dtype}")

                    # Print first 5 values each
                    print(f"  First 5 values: {dset[:5]}")

                    # Print units if available
                    # units = dset.attrs.get("units", "N/A")
                    # print(f"  Units: {units}")
        except Exception as exc:
            raise ValueError("An error occurred while reading the HDF5 file") from exc



#!/usr/bin/python3
r"""
    Add parameters and production tables from a simulation model repository to a new database.

    Generates a new database with all required collections.
    Follows the structure of the CTAO gitlab simulation model repository.

    This is an application for experts and should not be used by the general user.

    Command line arguments

    input_path (str, required)
        Path of local copy of model parameter repository.
    db_name (str, required)
        Name of new DB to be created.
    type (str, optional)
        Type of data to be uploaded to the DB. Options are: model_parameters, production_tables.

    Examples
    --------
    Upload model data repository to the DB:

    .. code-block:: console

        simtools-db-simulation-model-from-repository-to-db \
            --input_path /path/to/repository \
            --db_name new_db_name \
            --type model_parameters

    Upload production tables to the DB:

    .. code-block:: console

        simtools-db-simulation-model-from-repository-to-db \
            --input_path /path/to/repository \
            --db_name new_db_name \
            --type production_tables

"""

import logging
from pathlib import Path

import simtools.utils.general as gen
from simtools.configuration import configurator
from simtools.db import db_handler, db_model_upload


def _parse(label=None, description=None):
    """
    Parse command line configuration.

    Parameters
    ----------
    label : str
        Label describing application.
    description : str
        Description of application.

    Returns
    -------
    CommandLineParser
        Command line parser object.
    """
    config = configurator.Configurator(label=label, description=description)
    config.parser.add_argument(
        "--input_path",
        help="Path to simulation model repository.",
        type=Path,
        required=True,
    )
    config.parser.add_argument(
        "--db_name",
        help="Name of the new simulation model database to be created.",
        type=str.strip,
        required=True,
    )
    config.parser.add_argument(
        "--type",
        help="Type of data to be uploaded to the database.",
        type=str,
        required=False,
        default="model_parameters",
        choices=["model_parameters", "production_tables"],
    )

    args_dict, db_config = config.initialize(output=True, require_command_line=True, db_config=True)
    db_config["db_simulation_model"] = args_dict["db_name"]  # overwrite explicitly DB configuration
    return args_dict, db_config


def main():
    """Application main."""
    label = Path(__file__).stem
    args_dict, db_config = _parse(
        label, description="Add or update a model parameter database to the DB"
    )
    logger = logging.getLogger()
    logger.setLevel(gen.get_log_level_from_user(args_dict["log_level"]))

    db = db_handler.DatabaseHandler(mongo_db_config=db_config)

    if args_dict.get("type") == "model_parameters":
        db_model_upload.add_model_parameters_to_db(args_dict, db)
    elif args_dict.get("type") == "production_tables":
        db_model_upload.add_production_tables_to_db(args_dict, db)


if __name__ == "__main__":
    main()



#!/usr/bin/python3

"""
Plot array elements (array layout).

Plot an array layout and save it to file (e.g., pdf). Layouts are defined in the database,
or given as command line arguments (explicit listing or telescope list file). A list of input
files is also accepted.
Layouts can be plotted in ground or UTM coordinate systems.

Listing of array elements follows this logic:

* explicit listing: e.g., ``-array_element_list MSTN-01, MSTN05``
* listing of types: e.g, ``-array_element_list MSTN`` plots all telescopes of type MSTN.

A rotation angle in degrees allows to rotate the array before plotting.
The typical image formats (e.g., pdf, png, jpg) are allowed for the output figures.
If no ``figure_name`` is given as output, layouts are plotted in pdf and png format.

Example of a layout plot:

.. _plot_array_layout_plot:
.. image:: images/plot_array_layout_example.png
    :width: 49 %

Command line arguments
----------------------
figure_name : str
    File name for the output figure.
array_layout_file : str
    File (astropy table compatible) with a list of array elements.
array_layout_name : str
    Name of the layout array (e.g., test_layout, alpha, 4mst, etc.).
array_element_list : list
    List of array elements (e.g., telescopes) to plot (e.g., ``LSTN-01 LSTN-02 MSTN``).
coordinate_system : str, optional
    Coordinate system for the array layout (ground or utm).
rotate_angle : float, optional
    Angle to rotate the array before plotting (in degrees).
show_labels : bool, optional
    Shows the telescope labels in the plot.
axes_range : float, optional
    Range of the both axes in meters.
marker_scaling : float, optional.
    Scaling factor for plotting of array elements, optional.

Examples
--------
Plot layout with the name "test_layout":

.. code-block:: console

    simtools-plot-layout-array --figure_name northern_array_alpha
                               --array_layout_name test_layout


Plot layout with 2 LSTs and all northern MSTs in UTM coordinates:

.. code-block:: console

    simtools-plot-layout-array --array_element_list LSTN-01 LSTN-02 MSTN
                               --coordinate_system utm

Plot layout from a file with the list of telescopes:

.. code-block:: console

    simtools-plot-layout-array --array_element_list telescope_positions-test_layout.ecsv
"""

import logging
from pathlib import Path

import matplotlib as mpl
import matplotlib.pyplot as plt
from astropy import units as u

import simtools.utils.general as gen
from simtools.configuration import configurator
from simtools.io_operations import io_handler
from simtools.model.array_model import ArrayModel
from simtools.utils import names
from simtools.visualization.visualize import plot_array


def _parse(label, description, usage):
    """
    Parse command line configuration.

    Parameters
    ----------
    label : str
        Label describing the application.
    description : str
        Description of the application.
    usage : str
        Example on how to use the application.

    Returns
    -------
    CommandLineParser
        Command line parser object.
    """
    config = configurator.Configurator(label=label, description=description, usage=usage)

    config.parser.add_argument(
        "--figure_name",
        help="Name of the output figure to be saved into as a pdf.",
        type=str,
        required=False,
        default=None,
    )
    config.parser.add_argument(
        "--rotate_angle",
        help="Angle to rotate the array (in degrees).",
        type=str,
        required=False,
        default=None,
    )
    config.parser.add_argument(
        "--show_labels",
        help="Plot array element labels.",
        action="store_true",
        required=False,
        default=False,
    )
    config.parser.add_argument(
        "--marker_scaling",
        help="Scaling factor for the markers.",
        type=float,
        required=False,
        default=1.0,
    )
    config.parser.add_argument(
        "--coordinate_system",
        help="Coordinate system for the array layout.",
        type=str,
        required=False,
        default="ground",
        choices=["ground", "utm"],
    )
    config.parser.add_argument(
        "--axes_range",
        help="Range of the both axes in meters.",
        type=float,
        required=False,
        default=None,
    )
    return config.initialize(
        db_config=True, simulation_model=["site", "model_version", "layout", "layout_file"]
    )


def _get_site_from_telescope_list_name(telescope_list_file):
    """
    Get the site name from the telescope list file name.

    Parameters
    ----------
    telescope_list_file : str
        Telescope list file name.

    Returns
    -------
    str
        Site name.
    """
    for _site in names.site_names():
        if _site in str(telescope_list_file):
            return _site
    return None


def _get_list_of_plot_files(plot_file_name, output_dir):
    """
    Get list of output file names for plotting.

    Parameters
    ----------
    plot_file_name : str
        Name of the plot file.
    output_dir : str
        Output directory.

    Returns
    -------
    list
        List of output file names.

    Raises
    ------
    NameError
        If the file extension is not valid.
    """
    plot_file = output_dir.joinpath(plot_file_name)

    if len(plot_file.suffix) == 0:
        return [plot_file.with_suffix(f".{ext}") for ext in ["pdf", "png"]]

    allowed_extensions = [".jpeg", ".jpg", ".png", ".tiff", ".ps", ".pdf", ".bmp"]
    if plot_file.suffix in allowed_extensions:
        return [plot_file]
    msg = f"Extension in {plot_file} is not valid. Valid extensions are: {allowed_extensions}."
    raise NameError(msg)


def _get_plot_file_name(figure_name, layout_name, site, coordinate_system, rotate_angle):
    """
    Generate and return the file name for plots.

    Parameters
    ----------
    figure_name : str
        Figure name given through command line.
    layout_name : str
        Name of the layout.
    site : str
        Site name.
    coordinate_system : str
        Coordinate system for the array layout.
    rotate_angle : float
        Angle to rotate the array before plotting.

    Returns
    -------
    str
        Plot file name.
    """
    if figure_name is not None:
        return figure_name

    return (
        f"array_layout_{layout_name}_{site}_{coordinate_system}_"
        f"{round(rotate_angle.to(u.deg).value)!s}deg"
    )


def _layouts_from_array_layout_file(args_dict, db_config, rotate_angle):
    """
    Read array layout positions from file(s) and return a list of layouts.

    Parameters
    ----------
    args_dict : dict
        Dictionary with the command line arguments.
    db_config : dict
        Database configuration.
    rotate_angle : float
        Angle to rotate the array before plotting (in degrees).

    Returns
    -------
    list
        List of array layouts.
    """
    layouts = []
    telescope_files = args_dict["array_layout_file"]
    for one_file in telescope_files:
        site = (
            _get_site_from_telescope_list_name(one_file)
            if args_dict["site"] is None
            else args_dict["site"]
        )
        array_model = ArrayModel(
            mongo_db_config=db_config,
            model_version=args_dict["model_version"],
            site=site,
            array_elements=one_file,
        )
        layouts.append(
            {
                "array_elements": array_model.export_array_elements_as_table(),
                "plot_file_name": _get_plot_file_name(
                    args_dict["figure_name"],
                    (Path(one_file).name).split(".")[0],
                    site,
                    args_dict["coordinate_system"],
                    rotate_angle,
                ),
            }
        )
    return layouts


def _layouts_from_list(args_dict, db_config, rotate_angle):
    """
    Read positions for a list of array elements from the database and return a list of layouts.

    Parameters
    ----------
    args_dict : dict
        Dictionary with the command line arguments.
    db_config : dict
        Database configuration.
    rotate_angle : float
        Angle to rotate the array before plotting (in degrees).

    Returns
    -------
    list
        List of array layouts.
    """
    site = (
        names.get_site_from_array_element_name(args_dict["array_element_list"][0])
        if args_dict["site"] is None
        else args_dict["site"]
    )
    array_model = ArrayModel(
        mongo_db_config=db_config,
        model_version=args_dict["model_version"],
        site=site,
        array_elements=args_dict["array_element_list"],
    )
    return [
        {
            "array_elements": array_model.export_array_elements_as_table(
                coordinate_system=args_dict["coordinate_system"]
            ),
            "plot_file_name": _get_plot_file_name(
                args_dict["figure_name"],
                "list",
                site,
                args_dict["coordinate_system"],
                rotate_angle,
            ),
        }
    ]


def _layouts_from_db(args_dict, db_config, rotate_angle):
    """
    Read array elements and their positions from data base using the layout name.

    Parameters
    ----------
    args_dict : dict
        Dictionary with the command line arguments.
    db_config : dict
        Database configuration.
    rotate_angle : float
        Angle to rotate the array before plotting (in degrees).

    Returns
    -------
    list
        List of array layouts.
    """
    layouts = []
    array_model = ArrayModel(
        mongo_db_config=db_config,
        model_version=args_dict["model_version"],
        site=args_dict["site"],
        layout_name=args_dict["array_layout_name"],
    )
    layouts.append(
        {
            "array_elements": array_model.export_array_elements_as_table(
                coordinate_system=args_dict["coordinate_system"]
            ),
            "plot_file_name": _get_plot_file_name(
                figure_name=args_dict["figure_name"],
                layout_name=args_dict["array_layout_name"],
                site=args_dict["site"],
                coordinate_system=args_dict["coordinate_system"],
                rotate_angle=rotate_angle,
            ),
        }
    )
    return layouts


def main():
    """Plot array layout application."""
    label = Path(__file__).stem
    args_dict, db_config = _parse(
        label,
        "Plots array layout.",
        "python applications/plot_array_layout.py --array_layout_name test_layout",
    )
    logger = logging.getLogger()
    logger.setLevel(gen.get_log_level_from_user(args_dict["log_level"]))
    io_handler_instance = io_handler.IOHandler()

    rotate_angle = (
        0.0 * u.deg
        if args_dict["rotate_angle"] is None
        else float(args_dict["rotate_angle"]) * u.deg
    )

    layouts = []
    if args_dict["array_layout_name"] is not None:
        logger.info("Plotting array from layout array name.")
        layouts = _layouts_from_db(args_dict, db_config, rotate_angle)
    elif args_dict["array_layout_file"] is not None:
        logger.info("Plotting array from telescope list file.")
        layouts = _layouts_from_array_layout_file(args_dict, db_config, rotate_angle)
    elif args_dict["array_element_list"] is not None:
        logger.info("Plotting array from list of array elements.")
        layouts = _layouts_from_list(args_dict, db_config, rotate_angle)

    mpl.use("Agg")
    for layout in layouts:
        fig_out = plot_array(
            telescopes=layout["array_elements"],
            rotate_angle=rotate_angle,
            show_tel_label=args_dict["show_labels"],
            axes_range=args_dict["axes_range"],
            marker_scaling=args_dict["marker_scaling"],
        )
        _plot_files = _get_list_of_plot_files(
            layout["plot_file_name"],
            io_handler_instance.get_output_directory(label, sub_dir="application-plots"),
        )

        for file in _plot_files:
            logger.info(f"Saving figure as {file}")
            plt.savefig(file, bbox_inches="tight", dpi=400)
        fig_out.clf()
        plt.close()


if __name__ == "__main__":
    main()



"""MC model of a telescope."""

import logging
from pathlib import Path

import astropy.io.ascii
import numpy as np
from astropy.table import Table

import simtools.utils.general as gen
from simtools.model.camera import Camera
from simtools.model.mirrors import Mirrors
from simtools.model.model_parameter import InvalidModelParameterError, ModelParameter
from simtools.utils import names

__all__ = ["TelescopeModel"]


class TelescopeModel(ModelParameter):
    """
    TelescopeModel represents the MC model of an individual telescope.

    TelescopeModel contains parameter names and values for a specific telescope model.

    Parameters
    ----------
    site: str
        Site name (e.g., South or North).
    telescope_name: str
        Telescope name (ex. LSTN-01, LSTN-design, ...).
    mongo_db_config: dict
        MongoDB configuration.
    model_version: str
        Model version.
    label: str, optional
        Instance label. Important for output file naming.
    """

    def __init__(
        self,
        site: str,
        telescope_name: str,
        mongo_db_config: dict,
        model_version: str,
        label: str | None = None,
    ):
        """Initialize TelescopeModel."""
        super().__init__(
            site=site,
            array_element_name=telescope_name,
            mongo_db_config=mongo_db_config,
            model_version=model_version,
            db=None,
            label=label,
        )

        self._logger = logging.getLogger(__name__)
        self._logger.debug("Init TelescopeModel %s %s", site, telescope_name)

        self._single_mirror_list_file_paths = None
        self._mirrors = None
        self._camera = None

    @property
    def mirrors(self):
        """Load the mirror information if the class instance hasn't done it yet."""
        if self._mirrors is None:
            self._load_mirrors()
        return self._mirrors

    @property
    def camera(self):
        """Load the camera information if the class instance hasn't done it yet."""
        if self._camera is None:
            self._load_camera()
        return self._camera

    def export_single_mirror_list_file(self, mirror_number: int, set_focal_length_to_zero: bool):
        """
        Export a mirror list file with a single mirror in it.

        Parameters
        ----------
        mirror_number: int
            Number index of the mirror.
        set_focal_length_to_zero: bool
            Set the focal length to zero if True.
        """
        if mirror_number > self.mirrors.number_of_mirrors:
            logging.error("mirror_number > number_of_mirrors")
            return

        file_name = names.simtel_single_mirror_list_file_name(
            self.site, self.name, self.model_version, mirror_number, self.label
        )
        if self._single_mirror_list_file_paths is None:
            self._single_mirror_list_file_paths = {}
        self._single_mirror_list_file_paths[mirror_number] = self.config_file_directory.joinpath(
            file_name
        )

        # Using SimtelConfigWriter
        self._load_simtel_config_writer()
        self.simtel_config_writer.write_single_mirror_list_file(
            mirror_number,
            self.mirrors,
            self._single_mirror_list_file_paths[mirror_number],
            set_focal_length_to_zero,
        )

    def get_single_mirror_list_file(
        self, mirror_number: int, set_focal_length_to_zero: bool = False
    ):
        """
        Get the path to the single mirror list file.

        Parameters
        ----------
        mirror_number: int
            Mirror number.
        set_focal_length_to_zero: bool
            Flag to set the focal length to zero.

        Returns
        -------
        Path
            Path of the single mirror list file.
        """
        self.export_single_mirror_list_file(mirror_number, set_focal_length_to_zero)
        return self._single_mirror_list_file_paths[mirror_number]

    def _load_mirrors(self):
        """Load the attribute mirrors by creating a Mirrors object with the mirror list file."""
        mirror_list_file_name = self.get_parameter_value("mirror_list")
        self._logger.debug(f"Reading mirror list from {mirror_list_file_name}")
        try:
            mirror_list_file = gen.find_file(mirror_list_file_name, self.config_file_directory)
        except FileNotFoundError:
            mirror_list_file = gen.find_file(mirror_list_file_name, self.io_handler.model_path)
            self._logger.warning(
                "Mirror_list_file was not found in the config directory - "
                "Using the one found in the model_path"
            )
        except TypeError as exc:
            raise TypeError("Undefined mirror list") from exc

        self._mirrors = Mirrors(mirror_list_file, parameters=self._parameters)

    def get_telescope_effective_focal_length(
        self, unit: str = "m", return_focal_length_if_zero: bool = False
    ) -> float:
        """
        Return effective focal length.

        The function ensures backwards compatibility with older sim-telarray versions.

        Parameters
        ----------
        unit: str
            Unit of the effective focal length. Default is 'm'.
        return_focal_length_if_zero: bool
            If True, return the focal length if the effective focal length is 0.

        Returns
        -------
        float:
            Effective focal length.
        """
        try:
            eff_focal_length = self.get_parameter_value_with_unit("effective_focal_length")[0]
        except TypeError:
            eff_focal_length = self.get_parameter_value_with_unit("effective_focal_length")
        try:
            eff_focal_length = eff_focal_length.to(unit).value
        except AttributeError:
            eff_focal_length = 0.0
        if return_focal_length_if_zero and (eff_focal_length is None or eff_focal_length == 0.0):
            self._logger.warning("Using focal_length because effective_focal_length is 0")
            return self.get_parameter_value_with_unit("focal_length").to(unit).value
        return eff_focal_length

    def _load_camera(self):
        """Load camera attribute by creating a Camera object with the camera config file."""
        camera_config_file = self.get_parameter_value("camera_config_file")
        focal_length = self.get_telescope_effective_focal_length("cm", True)
        try:
            camera_config_file_path = gen.find_file(camera_config_file, self.config_file_directory)
        except TypeError as exc:
            self._logger.error(
                f"Camera config file {camera_config_file} or "
                f"config file directory ({self.config_file_directory}) is None"
            )
            raise TypeError from exc
        except FileNotFoundError:
            self._logger.warning(
                f"Camera config file {camera_config_file} not found in the config directory "
                f"{self.config_file_directory}. Using the one found in the model_path"
            )
            camera_config_file_path = gen.find_file(camera_config_file, self.io_handler.model_path)

        self._camera = Camera(
            telescope_model_name=self.name,
            camera_config_file=camera_config_file_path,
            focal_length=focal_length,
        )

    def is_file_2d(self, par: str) -> bool:
        """
        Check if the file referenced by par is a 2D table.

        Parameters
        ----------
        par: str
            Name of the parameter.

        Returns
        -------
        bool:
            True if the file is a 2D map type.
        """
        try:
            file_name = self.get_parameter_value(par)
        except KeyError:
            logging.error(f"Parameter {par} does not exist")
            return False

        file = self.config_file_directory.joinpath(file_name)
        with open(file, encoding="utf-8") as f:
            return "@RPOL@" in f.read()

    def read_two_dim_wavelength_angle(self, file_name: str | Path) -> dict:
        """
        Read a two dimensional distribution of wavelength and angle (z-axis can be anything).

        Return a dictionary with three arrays, wavelength, angles, z (can be transmission,
        reflectivity, etc.)

        Parameters
        ----------
        file_name: str or Path
            File assumed to be in the model directory.

        Returns
        -------
        dict:
            dict of three arrays, wavelength, degrees, z.
        """
        _file = self.config_file_directory.joinpath(file_name)
        self._logger.debug("Reading two dimensional distribution from %s", _file)
        line_to_start_from = 0
        with open(_file, encoding="utf-8") as f:
            for i_line, line in enumerate(f):
                if line.startswith("ANGLE"):
                    degrees = np.array(line.strip().split("=")[1].split(), dtype=np.float16)
                    line_to_start_from = i_line + 1
                    break  # The rest can be read with np.loadtxt

        _data = np.loadtxt(_file, skiprows=line_to_start_from)

        return {
            "Wavelength": _data[:, 0],
            "Angle": degrees,
            "z": np.array(_data[:, 1:]).T,
        }

    def get_on_axis_eff_optical_area(self) -> float:
        """Return the on-axis effective optical area (derived previously for this telescope)."""
        ray_tracing_data = astropy.io.ascii.read(
            self.config_file_directory.joinpath(self.get_parameter_value("optics_properties"))
        )
        if not np.isclose(ray_tracing_data["Off-axis angle"][0], 0):
            self._logger.error(
                f"No value for the on-axis effective optical area exists."
                f" The minimum off-axis angle is {ray_tracing_data['Off-axis angle'][0]}"
            )
            raise ValueError
        return ray_tracing_data["eff_area"][0]

    def read_incidence_angle_distribution(self, incidence_angle_dist_file: str) -> Table:
        """
        Read the incidence angle distribution from a file.

        Parameters
        ----------
        incidence_angle_dist_file: str
            File name of the incidence angle distribution

        Returns
        -------
        incidence_angle_dist: astropy.table.Table
            Instance of astropy.table.Table with the incidence angle distribution.
        """
        self._logger.debug(
            "Reading incidence angle distribution from %s",
            self.config_file_directory.joinpath(incidence_angle_dist_file),
        )
        return astropy.io.ascii.read(self.config_file_directory.joinpath(incidence_angle_dist_file))

    @staticmethod
    def calc_average_curve(curves: dict, incidence_angle_dist: Table) -> Table:
        """
        Calculate an average curve from a set of curves.

        The calculation uses weights the distribution of incidence angles provided in
        incidence_angle_dist.

        Parameters
        ----------
        curves: dict
            dict of with 3 "columns", Wavelength, Angle and z. The dictionary represents a two \
            dimensional distribution of wavelengths and angles with the z value being e.g., \
            reflectivity, transmission, etc.
        incidence_angle_dist: astropy.table.Table
            Instance of astropy.table.Table with the incidence angle distribution. The assumed \
            columns are "Incidence angle" and "Fraction".

        Returns
        -------
        average_curve: astropy.table.Table
            Instance of astropy.table.Table with the averaged curve.
        """
        weights = []
        for angle_now in curves["Angle"]:
            weights.append(
                incidence_angle_dist["Fraction"][
                    np.nanargmin(np.abs(angle_now - incidence_angle_dist["Incidence angle"].value))
                ]
            )

        return Table(
            [curves["Wavelength"], np.average(curves["z"], weights=weights, axis=0)],
            names=("Wavelength", "z"),
        )

    def export_table_to_model_directory(self, file_name: str, table: Table) -> str:
        """
        Write out a file with the provided table to the model directory.

        Parameters
        ----------
        file_name: str
            File name to write to.
        table: astropy.table.Table
            Instance of astropy.table.Table with the values to write to the file.

        Returns
        -------
        Path:
            Path to the file exported.
        """
        file_to_write_to = self.config_file_directory.joinpath(file_name)
        table.write(file_to_write_to, format="ascii.commented_header", overwrite=True)
        return file_to_write_to.absolute()

    def position(self, coordinate_system: str = "ground") -> list:
        """
        Get coordinates in the given system.

        Parameters
        ----------
        coordinate_system: str
            Coordinates system. Default is 'ground'.

        Returns
        -------
        list :
            List of telescope position in the requested coordinate system.

        Raises
        ------
        KeyError
            If the coordinate system is not found.
        """
        try:
            return self.get_parameter_value_with_unit(f"array_element_position_{coordinate_system}")
        except InvalidModelParameterError as exc:
            self._logger.error(f"Coordinate system {coordinate_system} not found.")
            raise exc
